# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""User configurable interactive editor styling settings loaders."""

from gettext import gettext as _

from dob_bright.termio import dob_in_user_warning

__all__ = (
    'load_obj_from_internal',
)


def load_obj_from_internal(
    controller,
    obj_name,
    internal,
    default_name=None,
    warn_tell_not_found=False,
):
    """"""

    def _load_obj_from_internal():
        loaded_obj = try_loading_internal(internal, obj_name)
        if loaded_obj is None:
            warn_tell_on_object_not_found(obj_name)
            if default_name:
                loaded_obj = try_loading_internal(internal, default_name)
                debug_loaded_default(loaded_obj)
        return loaded_obj

    def try_loading_internal(internal, obj_name):
        if not obj_name:
            return None
        # See if this is one of the basic baked-in styles/lexers/things.
        return getattr(internal, obj_name, None)

    def warn_tell_on_object_not_found(obj_name):
        if not obj_name or not warn_tell_not_found:
            return
        msg = _('No object from “{0}” named “{1}”.').format(
            internal.__name__, obj_name,
        )
        controller.client_logger.warning(msg)
        dob_in_user_warning(msg)  # Also blather to stdout.

    def debug_loaded_default(loaded_obj):
        controller.affirm(loaded_obj is not None)  # Because you specified default!
        controller.client_logger.debug(
            _('Loaded default object for “{0}”.'.format(obj_name))
        )

    return _load_obj_from_internal()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""A time tracker for the command line. Utilizing the power of nark."""

from gettext import gettext as _

import copy
import re
import sys
from datetime import datetime, time, timedelta

from nark.helpers.parsing import parse_factoid

from dob_bright.termio import (
    attr,
    barf_and_exit,
    bg,
    click_echo,
    echo_block_header,
    fg
)

from .fact_dressed import FactDressed
from .fix_times import (
    DEFAULT_SQUASH_SEP,
    mend_facts_times,
    must_complete_times,
    reduce_time_hint,
    then_extend_fact
)
from .parse_mistakes import prepare_log_msg

__all__ = (
    'parse_input',
)


def parse_input(controller, file_in=None, progress=None):
    """
    Import Facts from STDIN or a file.
    """

    # MAYBE/2018-05-16 00:11: (lb): Parse whole file before prompting.
    #                               Allow --yes to work here, too?
    #                               Not sure...
    #   For now, we'll just read one fact at a time.

    def _parse_input():
        raw_facts = parse_facts_input(file_in)
        new_facts = must_hydrate_facts(raw_facts)
        conflicts = must_complete_times(controller, new_facts, progress=progress)
        controller.affirm(not conflicts)  # (lb): 2019-01-19: This happen?
        repair_shoulder_fact_times(new_facts, raw_facts)
        must_not_conflict_existing(new_facts)
        return new_facts

    # ***

    def parse_facts_input(file_in):
        # A few reminders:
        # - On `dob import`, file_in is None -- but on `dob import -`,
        #   Click sets file_in to <stdin> _io.TextIOWrapper.
        # - On `echo '...' | dob import` → isatty() is False, but
        #   on `dob import | cat`        → isatty() is True.
        # - Dob will print help on `dob import`, but if the user runs
        #   `dob import -` then dob waits for input from the user, and
        #   processes it line by line (whenever the user his ENTER).
        #   This is expected behavior -- that's what the special `-`
        #   filename convention does in *nix -- but I thought I'm spell
        #   it out, in case it feels odd to new developers.
        input_f = file_or_stdin(file_in)
        raw_facts = parse_facts_from_stream(input_f)
        return raw_facts

    def file_or_stdin(file_in):
        if file_in is not None:
            return file_in
        return sys.stdin

    def parse_facts_from_stream(input_f):
        # Track current empty line run count, to know when to check if line
        # starts a new Fact. Keep at -1 until we've seen the first fact.
        bl_count = -1

        line_num = 0

        # Coalesce each Fact, line by line.
        current_fact_dict = None
        accumulated_fact = []
        unprocessed_facts = []

        progress and progress.click_echo_current_task(_('Parsing factoids...'))
        for line in input_f:
            line_num += 1
            (
                bl_count, processed,
            ) = gobble_blank_or_continued(line, accumulated_fact, bl_count)
            if processed:
                continue
            fact_dict = gobble_if_not_new_fact(line, accumulated_fact, bl_count)
            if fact_dict is None:
                bl_count = 0
                continue
            fact_dict['parsed_source.line_num'] = line_num
            fact_dict['parsed_source.line_raw'] = line
            if not accumulated_fact:
                # First Fact.
                assert current_fact_dict is None
                assert not accumulated_fact
                current_fact_dict = fact_dict
                accumulated_fact.append(line)
                bl_count = 0
                continue
            else:
                assert current_fact_dict is not None
                assert accumulated_fact

            # Woot, woot! We parsed a complete Fact.
            if current_fact_dict:
                unprocessed_facts.append((current_fact_dict, accumulated_fact,))
            current_fact_dict = fact_dict
            accumulated_fact = [line, ]
            bl_count = 0

        # end: for

        if accumulated_fact:
            unprocessed_facts.append((current_fact_dict, accumulated_fact,))
        else:
            msg = _('What is this, an empty file?')
            barf_and_exit(msg)

        return unprocessed_facts

    def gobble_blank_or_continued(line, accumulated_fact, blank_line_count):
        processed = False
        if blank_line_count < 0:
            # Start of file.
            assert blank_line_count == -1
            if not line.strip():
                controller.client_logger.debug(_('Skip premature blank line'))
                # Skip blank lines pre-Facts.
                processed = True
            # else, we found the first non-blank line. We now
            #   expect to find the date:time & meta, or death.
        elif not line.strip():  # remove trailing newline
            controller.client_logger.debug(_('- Blank line in desc.'))
            accumulated_fact.append(line)
            blank_line_count += 1
            processed = True
        elif blank_line_count == 0:
            controller.client_logger.debug(_('- Part of desc.:\n') + line.strip())
            accumulated_fact.append(line)
            processed = True
        elif re.match(r'^\s', line):
            controller.client_logger.debug(_('- Leading whitesp.:\n') + line.strip())
            accumulated_fact.append(line)
            processed = True
        # else, any content that follows blank line(s) is either:
        # (1) more content; (2) a new Fact; or (3) Fact separator.
        return (blank_line_count, processed)

    def gobble_if_not_new_fact(line, accumulated_fact, blank_line_count):
        fact_dict = dissect_meta_line(line)
        if fact_dict is None:
            # If not parsed, and first line we've seen, die.
            missing_fact_must_come_early(line, blank_line_count)
            # else, more content, or a Fact separator.
            controller.client_logger.debug(_('- More desc.: ') + line.strip())
            accumulated_fact.append(line)
        return fact_dict

    def missing_fact_must_come_early(line, blank_line_count):
        if blank_line_count == -1:
            msg = (
                '{}: {}{}{}{}{}\n'
                .format(
                    _('The first nonempty line is not a recognized Fact.'),
                    bg('white'), fg('black'), attr('underlined'),
                    line.strip(),
                    attr('reset'),
                )
            )
            barf_and_exit(msg)

    # FIXME/2019-01-21: Document all the different usage, both in test, and README.
    # E.g., "then: I did this" should be same as "at +0: I did this"
    #   but if you specify start you don't need colon
    #       "then 2019-01-21 23:47 I did this" should also work...
    #   FIXME: Parse "then" expecting 1+ datetimes, or "then:" expecting none.
    #   FIXME: Parse "still" expecting 1+ datetimes, or "still:" expecting none.
    # FIXME: Three tests:
    #           ``still <time-spec> <desc>``
    #       vs. ``still: <desc>``
    #       vs. ``still blah``
    RE_TIME_HINT = re.compile(
        # SYNC_ME: RE_TIME_HINT, TIME_HINT_MAP, and @generate_add_fact_command's.
        r'^('
            # Skipping: Doesn't make sense: '(?P<verify_none>on|now)'
            # MAYBE/2019-01-22: Is "between" okay here?
            #   We don't have a Click alias for it, so
            #   there's a `dob from` command, but not `dob between`,
            #   so maybe we want to remove "between" from here.
            '(?P<verify_both>from|between)'
            '|(?P<verify_start>at)'  # noqa: E131
            '|(?P<verify_end>to|until)'
            '|(?P<verify_then_none>then:)'
            '|(?P<verify_then_some>then)'
            '|(?P<verify_still_none>still:)'
            '|(?P<verify_still_some>still)'
            # NOTE: Require colon postfix for options w/o time component.
            # NOTE: 'now' would be confusing and conflict with other usage,
            #       (at least I think it would?). E.g., do not do this:
            #         '|(?P<verify_after>after:|since:|next:|now:)'
            '|(?P<verify_after>after:|since:|next:)'
        ' )',  # NOTE The SPACE CHARACTER following THE DIRECTIVE!
        re.IGNORECASE,
    )

    def dissect_meta_line(line):
        sussed_hint, time_hint, line = suss_the_time_format(line)
        fact_dict, err = run_parser(line, time_hint)
        if not fact_dict['start'] and not fact_dict['end']:
            controller.client_logger.debug(_('unmeta: {}'.format(err)))
            fact_dict = None
        else:
            # The user could specify, e.g., just a single datetime (followed
            # by a description), and we can fill in the other datetime and
            # ask the user for more details about the fact.
            #   Not always True:  assert not err
            controller.client_logger.debug(_(
                # Including the description is too verbose:
                #   'new fact_dict: {}'.format(fact_dict)
                # so let's make a complicated dictionary comprehension instead.
                'new fact_dict: {}'.format(
                    {
                        key: val if key != 'description'
                        else val[:10] + ((len(val) > 10) and '...' or '')
                        for key, val in fact_dict.items()
                    }
                )
            ))
        return fact_dict

    def suss_the_time_format(line):
        sussed_hint = ''
        match = RE_TIME_HINT.match(line)
        if match is not None:
            for hint, matched in match.groupdict().items():
                if matched is not None:
                    assert not sussed_hint
                    sussed_hint = hint
                    # Remove the time hint prefix.
                    line = RE_TIME_HINT.sub('', line).lstrip()
        time_hint = sussed_hint or 'verify_start'
        controller.client_logger.debug(_(
            'time_hint: {}{}'.format(
                time_hint, ' [sussed]' if sussed_hint else ' [default]',
            )
        ))
        return sussed_hint, time_hint, line

    def run_parser(line, time_hint):
        # Parse the line as if it were a new fact. And don't strip the line.
        # Parser expects metadata to be separated from description (so leave
        # the newline). And we can be strict and require that the date data
        # starts the line.

        # Per dob-insert commands, parser expects iterable.
        factoid = (line,)

        use_hint = reduce_time_hint(time_hint)

        fact_dict, err = parse_factoid(
            factoid=factoid,
            time_hint=use_hint,
            # MEH: (lb): Should we leave hash_stamps='#@' ?
            #   I sorta like using the proper tag symbol
            #   when not worried about shell interpolation.
            hash_stamps='#',
            # We'll handle errors ourselves, later, in bulk, either
            # `--ask`'ing the user for more Fact details, or barfing
            # all the errors and exiting.
            lenient=True,
        )

        fact_dict_set_time_hint(fact_dict, time_hint)

        return fact_dict, err

    # FIXME/DRY: See create.py/transcode.py (places that use "+0").
    def fact_dict_set_time_hint(fact_dict, time_hint):
        fact_dict['time_hint'] = time_hint
        if time_hint in ('verify_after', 'verify_then_none', 'verify_still_none'):
            assert not fact_dict['start'] and not fact_dict['end']
            # (lb): How's this for a hack!?
            fact_dict['start'] = "+0"

    # ***

    def must_hydrate_facts(raw_facts):
        new_facts, hydrate_errs = hydrate_facts(raw_facts)
        must_hydrated_all_facts(hydrate_errs)
        return new_facts

    def hydrate_facts(raw_facts):
        new_facts = []
        hydrate_errs = []
        temp_id = -1
        progress and progress.click_echo_current_task(_('Hydrating Facts...'))
        for fact_dict, accumulated_fact in raw_facts:
            add_hydration_warnings(fact_dict, hydrate_errs)
            hydrate_description(fact_dict, accumulated_fact)
            new_fact, err_msg = create_fact_from_parsed_dict(fact_dict)
            if new_fact:
                assert not err_msg
                time_hint = fact_dict['time_hint']
                temp_id = add_new_fact(new_fact, time_hint, temp_id, new_facts)
            else:
                assert not new_fact
                hydrate_errs.append(err_msg)
        return new_facts, hydrate_errs

    def add_hydration_warnings(fact_dict, hydrate_errs):
        if not fact_dict['warnings']:
            return
        fact_warnings = '\n  '.join(fact_dict['warnings'])
        display_dict = copy.copy(fact_dict)
        del display_dict['warnings']
        err_msg = prepare_log_msg(display_dict, fact_warnings)
        hydrate_errs.append(err_msg)

    # Horizontal rule separator matches same character repeated at least thrice.
    # FIXME/2018-05-18: (lb): Document: HR is any repeated one of -, =, #, |.
    FACT_SEP_HR = re.compile(r'^([-=#|])\1{2}\1*$')

    def hydrate_description(fact_dict, accumulated_fact):
        # The first entry is the meta line, so ignore it.
        desc_lines = accumulated_fact[1:]
        # However, the user is allowed to start the description
        # on the meta line.
        if fact_dict['description']:
            # Don't forget the newline -- it got culled (strip()ped) on parse.
            desc_lines.insert(0, fact_dict['description'] + "\n")

        cull_factless_fact_separator(desc_lines)

        # desc_lines are already newlined.
        desc = ''.join(desc_lines).strip()

        fact_dict['description'] = desc

    def cull_factless_fact_separator(desc_lines):
        # To make import file more readable, user can add
        # separator line between facts. Cull it if found.
        # MAYBE/2018-05-18: (lb): Make this operation optional?
        while len(desc_lines) > 0 and not desc_lines[-1].strip():
            desc_lines.pop()

        if (
            len(desc_lines) > 1
            and not desc_lines[-2].strip()
            and FACT_SEP_HR.match(desc_lines[-1])
        ):
            # Remove optional horizontal rule, which user
            # can use to visual separate lines in import file.
            desc_lines.pop()

        return desc_lines

    def create_fact_from_parsed_dict(fact_dict):
        new_fact = None
        err_msg = None
        try:
            new_fact = FactDressed.create_from_parsed_fact(
                fact_dict,
                lenient=True,
                line_num=fact_dict['parsed_source.line_num'],
                line_raw=fact_dict['parsed_source.line_raw'],
            )
        except ValueError as err:
            err_msg = prepare_log_msg(fact_dict, str(err))
        # (lb): Returning a tuple that smells like Golang: (fact, err, ).
        return new_fact, err_msg

    def add_new_fact(new_fact, time_hint, temp_id, new_facts):
        if new_fact is None:
            return temp_id
        add_facts = maybe_squash_extend_prev(new_fact, time_hint, new_facts)
        for add_fact in add_facts:
            temp_id = add_new_fact_maybe(add_fact, temp_id, new_facts)
        return temp_id

    def maybe_squash_extend_prev(new_fact, time_hint, new_facts):
        if time_hint not in [
            'verify_end',
            'verify_then_none',
            'verify_then_some',
            'verify_still_none',
            'verify_still_some',
        ]:
            return [new_fact]
        return squash_extend_if_prev(new_fact, time_hint, new_facts)

    def squash_extend_if_prev(new_fact, time_hint, new_facts):
        prev_fact = new_facts[-1] if new_facts else None
        new_new_facts = squend_check_noop_or_extend(prev_fact, new_fact, time_hint)
        if new_new_facts is not None:
            return new_new_facts
        return squend_squash_or_extend(prev_fact, new_fact, time_hint, new_facts)

    def squend_check_noop_or_extend(prev_fact, new_fact, time_hint):
        if prev_fact is None:
            # First fact in import. Will check store later for boundary fact.
            return [new_fact]
        elif prev_fact.end is not None:
            return maybe_set_start_maybe_extend_prev(
                prev_fact, new_fact, time_hint,
            )
        elif prev_fact.start is None:
            return [new_fact]
        elif new_fact.start and new_fact.end:
            # This fact is a `then ... to`, so really just a `from/between`.
            return [new_fact]
        # E.g., prev_fact is at-fact, with no end; and new_fact is to-fact,
        # with no start. Return None, and caller will squend_squash_or_extend.
        assert not new_fact.deleted
        return None

    def squend_squash_or_extend(prev_fact, new_fact, time_hint, new_facts):
        squash_facts_maybe(new_fact, prev_fact)
        if not new_fact.deleted:
            return [new_fact]

        prev_or_new_fact = maybe_extend_fact(controller, prev_fact, time_hint)
        # Return None is prev fact was extended.
        if prev_or_new_fact is not new_fact:
            assert prev_or_new_fact is new_facts[-1]
        return [prev_or_new_fact] if prev_or_new_fact is new_fact else []

    def maybe_set_start_maybe_extend_prev(prev_fact, new_fact, time_hint):
        if time_hint in [
            'verify_then_none',
            'verify_then_some',
            'verify_still_none',
            'verify_still_some',
        ]:
            extended_fact = then_extend_fact(controller, prev_fact)
            extended_fact.end = new_fact.start
            if time_hint in ['verify_still_none', 'verify_still_some']:
                new_fact.activity = prev_fact.activity
                new_fact.tags = list(prev_fact.tags)
            return [extended_fact, new_fact]
        else:
            assert time_hint == 'verify_end'  # ``to``
            assert new_fact.start is None
            # Fix it in post, er, fix_delta_time_relative.
            new_fact.start = "+0"
            return [new_fact]

    def squash_facts_maybe(new_fact, prev_fact):
        assert new_fact.pk is None
        # 2018-06-30: (lb): Ug. Why does this squash logic seem so forced?
        new_fact.pk = -1  # To appease squash().
        prev_fact.squash(new_fact, DEFAULT_SQUASH_SEP)
        new_fact.pk = None

    def maybe_extend_fact(controller, new_fact, time_hint):
        if time_hint not in ['verify_then_none', 'verify_then_some']:
            return new_fact
        return then_extend_fact(controller, new_fact)

    def add_new_fact_maybe(new_fact, temp_id, new_facts):
        if not new_fact or new_fact.deleted:
            return temp_id
        new_fact.pk = temp_id
        new_facts.append(new_fact)
        temp_id -= 1
        return temp_id

    # ***

    def must_hydrated_all_facts(hydrate_errs):
        if not hydrate_errs:
            return
        for err_msg in hydrate_errs:
            click_echo()
            click_echo(err_msg)
            click_echo()
        msg = (_(
            'Please fix your import data and try again.'
            ' Scroll up for details.'
        ))
        barf_and_exit(msg)

    # ***

    def repair_shoulder_fact_times(new_facts, raw_facts):
        # If user did not specify first fact start time, and/or last fact's
        # end time, fix_times.must_complete_times will have assigned times.
        # Note that there are a number of places where two facts may be merged:
        # - Earlier during this import task, under add_new_fact(), squash and
        #   extend code is used.
        # - Here, but just on the first and final import facts, to sew them
        #   into the fabric of the facts in the store.
        # - In many of the fix_times functions:
        #     unite_and_stretch, insert_forcefully, and resolve_overlapping.
        # (lb): Maybe someday we can combine this code better. The way different
        #   features were grown, however, resulted in squash code in a few places.

        if not new_facts[0].start:
            # User did not specify first fact's start, and we did not find
            # an antecedent fact in the data store, so, what? Set earliest
            # time ever? Clip to midnight that day?
            new_facts[0].start = datetime.combine(new_facts[0].end, time.min)
            if new_facts[0].start == new_facts[0].end:
                new_facts[0].start = new_facts[0].end - timedelta(days=1)
        elif not raw_facts[0][0]['start']:
            new_facts[0].start = None

        controller.affirm(new_facts[-1].end)
        if not raw_facts[-1][0]['end']:
            new_facts[-1].end = None

    # ***

    def must_not_conflict_existing(new_facts):
        # (lb): Yuck. Sorry about this. Totally polluting what was a small
        # function with lots of progress-output overhead.
        task_descrip = _('Verifying times nonconflicting')
        if progress is not None:
            term_width, dot_count, fact_sep = progress.start_crude_progressor(
                task_descrip,
            )

        could_be_more = fix_range_conflicts_easy(new_facts)

        all_conflicts = []
        for idx, fact in enumerate(new_facts):
            if progress is not None:
                term_width, dot_count, fact_sep = progress.step_crude_progressor(
                    task_descrip, term_width, dot_count, fact_sep,
                )

            if idx == len(new_facts) - 1:
                time_hint = 'verify_last'
            elif idx == 0 and not fact.start:
                time_hint = 'verify_end'
            else:
                time_hint = 'verify_both'

            conflicts = mend_facts_times(
                controller,
                fact,
                time_hint=time_hint,
                skip_store=not could_be_more,
            )
            assert not fact.deleted  # Only on squash, which shouldn't happen.

            if conflicts:
                all_conflicts.append((fact, conflicts,))

        progress and progress.click_echo_current_task('')

        # Handling anything other than the ongoing (active) fact seems
        # beyond the scope of the import function. dob is not a conflict
        # resolution application!
        barf_on_overlapping_facts_old(all_conflicts)

    def fix_range_conflicts_easy(new_facts):
        """
        Check time range of facts being imported against store.
        Returns True if time range overlaps with any existing facts.
        The active fact overlaps if it starts during the time range.
        """
        def _fix_range_conflicts_easy():
            first_time = new_facts[0].start or new_facts[0].end
            final_time = new_facts[-1].end or new_facts[-1].start
            # There's a compelling function, controller.facts.strictly_during,
            # but it ignores overlapping shoulder facts (and the active fact).
            # Check instead manually against the fact before the last time,
            # and also the fact after the first time.
            ante_fact = controller.facts.antecedent(ref_time=final_time)
            seqt_fact = controller.facts.subsequent(ref_time=first_time)
            check_all = False
            check_all = easy_fix_final(check_all, seqt_fact, final_time)
            check_all = easy_fix_first(check_all, ante_fact, first_time)
            return check_all

        def easy_fix_final(check_all, seqt_fact, final_time):
            if not seqt_fact:
                return check_all
            if seqt_fact.start < final_time:
                # Fact found in store ends after first_time,
                #  and starts before final_time.
                return True
            if new_facts[-1].end:
                # The final import fact ends on or before
                #  the start of the after-fact.
                return check_all
            # If here, the final import fact is endless.
            # Obvious path is to end it using start of first fact after.
            if new_facts[-1].start == seqt_fact.start:
                # The ongoing final import fact starts at same time as final store fact.
                if not seqt_fact.end:
                    # To test this path: have ongoing fact in import with same start as
                    # ongoing (active) fact in the store.
                    # (lb): We could end the existing fact, making it momentaneous,
                    # but that seems extreme; feels more natural to combine the 2
                    # facts. -- Alternatively, we could toss an error. But I could
                    # see just combining the 2 facts, seems like a pretty rare case
                    # anyway.
                    seqt_fact.squash(new_facts[-1], DEFAULT_SQUASH_SEP)
                    new_facts[-1] = seqt_fact
                    return check_all
                if not controller.config['time.allow_momentaneous']:
                    # User is not allowing momentaneous Facts, so disallow.
                    # Return True and we'll run thorough conflict check
                    # (and exit on error instead of continuing import).
                    return True
                # The final import fact is endless, but it conflicts with a
                # fact in the store. And the user is down with momentaneous
                # so crown it as such.
                # FIXME/2019-01-20 01:09: TESTME: Momentaneous feature.
                #  (I think it works in db and code, but no way for user
                #  to view in the UI. So they're essentially really well
                #  hidden. Which is a design feature, maybe?)
            new_facts[-1].end = seqt_fact.start
            return check_all

        def easy_fix_first(check_all, ante_fact, first_time):
            if not ante_fact:
                return check_all
            if ante_fact.end:
                if not new_facts[0].start:
                    # I.e., a to-Fact is 1st Fact in import,
                    #   and final Fact in store is closed.
                    new_facts[0].start = ante_fact.end
                if ante_fact.end > first_time:
                    # Fact from store starts before final_time,
                    #  and ends after first_time.
                    return True
                # else, first fact found before final_time
                #        ends before first_time.
                return check_all
            # If here, it's the ongoing (active) fact.
            ongoing = controller.facts.endless()
            controller.affirm(ante_fact == ongoing[0])
            if new_facts[0].start:
                ante_fact.end = new_facts[0].start
                new_facts.insert(0, ante_fact)
            else:
                # Squash store's ongoing (active) and import's first.
                ante_fact.squash(new_facts[0], DEFAULT_SQUASH_SEP)
                new_facts[0] = ante_fact

            return check_all

        return _fix_range_conflicts_easy()

    def barf_on_overlapping_facts_old(conflicts):
        if not conflicts:
            return

        for fact, resolved_edits in conflicts:
            for edited_fact, original in resolved_edits:
                echo_saved_fact_conflict(fact, edited_fact, original)
        msg = _(
            'One or more new Facts would conflict with existing Facts.'
            ' This Is Not Allowed'
        )
        barf_and_exit(msg)

    def echo_saved_fact_conflict(fact, edited_fact, original):
        echo_block_header(_('Saved Fact Datetime Conflict!'))
        click_echo()
        click_echo(_('  Fact being imported:'))
        click_echo(_('  ───────────────────'))
        click_echo(fact.friendly_diff(fact, truncate=True))
        click_echo()
        click_echo(_('  Impact on saved Fact:'))
        click_echo(_('  ────────────────────'))
        click_echo(original.friendly_diff(edited_fact))
        click_echo()

    # ***

    return _parse_input()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from datetime import timedelta

__all__ = (
    'FactsManager_JumpTime',
)


class FactsManager_JumpTime(object):
    """"""
    def __init__(self, *args, **kwargs):
        super(FactsManager_JumpTime, self).__init__()

        self._jump_time_reference = None

    # ***

    @property
    def jump_time_reference(self):
        self.debug('get: {}'.format(
            self._jump_time_reference or '{} (reset)'.format(self.curr_fact.start)
        ))
        if not self._jump_time_reference:
            if not self.curr_fact.end:
                # If user is looking at Active Fact, and they want to, e.g.,
                # see Fact 1 day ago, use now time, not start of Fact, or
                # we might show a Fact before one they'd naturally expect.
                time_ref = self.controller.now
            else:
                # (lb): For any other Fact, I dunno, split the diff.
                half_delta = (self.curr_fact.end - self.curr_fact.start) / 2
                time_ref = self.curr_fact.start + half_delta
            self.jump_time_reference = time_ref
        return self._jump_time_reference

    @jump_time_reference.setter
    def jump_time_reference(self, jump_time_reference):
        self.debug('set: {}'.format(jump_time_reference))
        self._jump_time_reference = jump_time_reference

    # ***

    def jump_day_dec(self, days=1):
        prev_day = self.jump_time_reference - timedelta(days=days)
        prev_fact = self.jump_to_fact_nearest(until_time=prev_day)
        if prev_fact is None:
            prev_fact = self.jump_to_oldest_fact(reason='day-dec')
        return prev_fact

    def jump_day_inc(self, days=1):
        next_day = self.jump_time_reference + timedelta(days=days)
        next_fact = self.jump_to_fact_nearest(since_time=next_day)
        if next_fact is None:
            next_fact = self.jump_to_latest_fact(reason='day-inc')
        return next_fact




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from contextlib import contextmanager

from sortedcontainers import SortedKeyList

from .facts_mgr_fact_dec import FactsManager_FactDec
from .facts_mgr_fact_inc import FactsManager_FactInc
from .facts_mgr_gap import FactsManager_Gap
from .facts_mgr_jump import FactsManager_Jump
from .facts_mgr_jump_time import FactsManager_JumpTime
from .facts_mgr_rift import FactsManager_Rift
from .facts_mgr_rift_dec import FactsManager_RiftDec
from .facts_mgr_rift_inc import FactsManager_RiftInc
from .group_chained import GroupChained

__all__ = (
    'FactsManager',
)


class FactsManager(
    FactsManager_FactDec,
    FactsManager_FactInc,
    FactsManager_Gap,
    FactsManager_Jump,
    FactsManager_JumpTime,
    FactsManager_Rift,
    FactsManager_RiftDec,
    FactsManager_RiftInc,
):
    """"""

    # ***

    def __init__(self, controller, on_insert_fact, on_jumped_fact, *args, **kwargs):
        super(FactsManager, self).__init__(controller, *args, **kwargs)

        self.controller = controller
        self.on_insert_fact = on_insert_fact
        self.on_jumped_fact = on_jumped_fact
        self.debug = controller.client_logger.debug
        self.groups = self.sorted_contiguous_facts_list()
        self.by_pk = {}
        self.last_fact_pk = 0
        self._curr_fact = None
        self.curr_group = None
        self.curr_index = None

    def sorted_contiguous_facts_list(self):
        sorted_contiguous_facts_list = SortedKeyList(
            key=lambda group_chained: (group_chained.sorty_times),
        )
        return sorted_contiguous_facts_list

    # ***

    @property
    def curr_fact(self):
        return self._curr_fact

    @curr_fact.setter
    def curr_fact(self, curr_fact):
        if self.curr_fact is curr_fact:
            return
        group, index = self.locate_fact(curr_fact)
        self._curr_fact = curr_fact
        self.curr_group = group
        self.curr_index = index
        # (lb): 2019-01-21: This seems unnecessary, but why not.
        #   "In the name of coveragggggggggge!!!!!!!"
        # 2019-01-23: Hahaha, it fired on self.curr_group.time_since not
        #   having been extended after collapse_group! Hooray, affirm usage!
        self.controller.affirm(self.curr_group.contains_fact_time([curr_fact]))

    def locate_fact(self, some_fact):
        inserts_at = self.groups.bisect_key_left(some_fact.sorty_times)
        if (
            (inserts_at < len(self.groups))
            and (self.groups[inserts_at])
            # (lb): 2019-01-31: We had been comparing PKs here:
            #   and (self.groups[inserts_at][0].pk == some_fact.pk)
            # but I'm pretty sure we want to check the time window.
            #  (Such glaring confidence, I know!)
            and (self.groups[inserts_at][0].sorty_times == some_fact.sorty_times)
        ):
            group = self.groups[inserts_at]
        else:
            group = self.groups[inserts_at - 1]
        # The index raises ValueError if the fact is not in the group.
        index = group.index(some_fact)
        return group, index

    def locate_wired(self, ref_fact):
        try:
            group, index = self.locate_fact(ref_fact)
        except KeyError:
            return None
        else:
            return group[index]

    # ***

    def __getitem__(self, key):
        if key == 0:
            return self.groups[0][0]
        elif key == -1:
            return self.groups[-1][-1]
        raise TypeError(
            "'{0}' object is not really subscriptable".format(type(self))
        )

    def __len__(self):
        return sum([len(group) for group in self.groups])

    @property
    def debug__str(self):
        return '  ' + '\n  '.join(
            ['#{:3}: {}'.format(idx, grp) for idx, grp in enumerate(self.groups)]
        )

    # ***

    def add_facts(self, facts):
        if not facts:
            return

        grouped_facts = []
        for fact in facts:
            self.controller.affirm(fact.pk not in self.by_pk.keys())
            self.by_pk[fact.pk] = fact
            grouped_facts.append(fact)
            # For creating new Facts.
            if fact.unstored:
                self.last_fact_pk = min(self.last_fact_pk, fact.pk)

        # FIXME/2019-12-06: (lb): Just testing. Remove affirm arg. later.
        # group = GroupChained(grouped_facts)
        group = GroupChained(grouped_facts, affirm=self.controller.affirm)
        self.groups.add(group)

        self.logger_debug_groups('add_facts', group=group)

    def claim_time_span(self, since, until):
        owning_group = None

        sorty_times = (since, since)

        inserts_at = self.groups.bisect_key_left(sorty_times)
        # Because bisect_key_left, either the start of the indexed
        # group matches, or the time falls before the group indexed.
        if inserts_at < len(self.groups):
            if since == self.groups[inserts_at].time_since:
                owning_group = self.groups[inserts_at]
            else:
                self.controller.affirm(since < self.groups[inserts_at].time_since)
        if (owning_group is None) and (inserts_at > 0):
            before_it = inserts_at - 1
            self.controller.affirm(since > self.groups[before_it].time_since)
            if since <= self.groups[before_it].time_until:
                owning_group = self.groups[before_it]
            # else, since is between 2 groups; until might extend
            # past 2nd of the groups, which we'll check for later.

        if inserts_at == 0:
            # Because bisect_key_left, leftmost group
            # only matches if start exactly matches.
            if since == self.groups[0][0].start:
                owning_group = self.groups[0]

        else:
            # Time span precedes all existing groups' spans.
            self.controller.affirm(since < self.groups[0].time_since)

        if owning_group is None:
            owning_group = GroupChained()

        with self.fact_group_rekeyed(owning_group):
            owning_group.claim_time_span(since, until)

    # ***

    def apply_edits(self, edit_facts, last_edits):
        group, _index = self.locate_fact(last_edits[0])

        with self.fact_group_rekeyed(group):

            # Clear time windows of edited facts, as times may have changed.
            for last_edit in last_edits:
                last_index = group.index(last_edit)
                group_fact = group.pop(last_index)
                # 2020-04-09: (lb): I had this affirm here:
                #   self.controller.affirm(group_fact == last_edit)
                # which meant to say that the Fact in the Fact Manager
                # group matches the most recent Fact edit, i.e., the
                # group fact has not been updated yet.
                # However, on time edit, the apply_edit_time_start/end
                # methods update the editable fact, but the editable
                # facts are also part of the Facts Manager groups. So
                # while group_fact.pk == last_edit.pk, other attrs might
                # now differ.
                self.controller.affirm(
                    (group_fact == last_edit) or (group_fact in edit_facts)
                )

                if group_fact.has_prev_fact:
                    group_fact.prev_fact.next_fact = None
                if group_fact.has_next_fact:
                    group_fact.next_fact.prev_fact = None

                group_fact.prev_fact = None
                group_fact.next_fact = None

                del self.by_pk[group_fact.pk]

            for edit_fact in edit_facts:
                # Rather than try to rewire the Facts, e.g., by calling
                #   self.new_fact_wire_links(edit_fact)
                # leave the Facts unwired, and let the normal _inc/_dec
                # navigation methods fix the wiring.
                edit_fact.next_fact = None
                edit_fact.prev_fact = None
                group.add(edit_fact)
                self.by_pk[edit_fact.pk] = edit_fact

    # ***

    @contextmanager
    def fact_group_rekeyed(self, group=None, group_index=None):
        # Ensures that self.groups._keys is up to date.

        # The group's facts order should not change, but the group's
        # key might change, if some_fact is being prepended to the
        # group, because self.curr_group.facts[0].sorty_times.
        # As such, remove and re-add the group, so that SortedKeyList
        # can update, e.g., self.groups._maxes is set when a group is
        # updated, so really a group is invariant once it's added to
        # the sorted list. (If we didn't re-add the group, things happen,
        # like, self.groups.index(self.curr_group) will not find the
        # group if its sorty_times < _maxes.)
        group = group or self.curr_group
        if group_index is None:
            # MAYBE/2019-01-21: self.groups.index will raise ValueError
            #  if you edited the first or final fact while the group is
            #  still part of the sorted_contiguous_facts_list() container.
            # This is not, like, a moral issue, or anything, but for the
            #  sake of our code, anything that changes the group should
            #  be aware of this.
            # However, if we find that maintaining the code as such starts
            #  to become painful -- should this dance become more difficult
            #  -- we could fallback and walk self.groups for an object match.
            try:
                group_index = self.groups.index(group)
            except ValueError:
                # MAYBE/2019-01-21: See long comment from a few lines back.
                #  Look for exact group object match (i.e., instead of using
                #  sorty_times value). This is because groups.index uses
                #  _maxes and compares key values, ignoring object identity.
                #   and group.sorty_times changes based on its facts, and
                #   when we extend the time window. (lb): Long comment....
                self.controller.affirm(False)  # Unexpected path, but may work:
                for idx, grp in enumerate(self.groups):
                    if grp is group:
                        group_index = idx
                        break
                if group_index is None:
                    raise

        if group_index is not None:
            # NOTE: Use pop(), specifying an index, rather than remove(),
            #       which uses a key value, because sorty_times might already
            #       be invalid.
            self.groups.pop(group_index)

        yield

        self.groups.add(group)

        self.logger_debug_groups('fact_group_rekeyed', group=group)

        # Caller is responsible for wiring prev/next references.

    def logger_debug_groups(self, whence='', group=None):
        group = group or self.curr_group
        self.debug(
            '{}\n- group.sorty_times: {}\n-    groups._maxes: {}'.format(
                whence,
                group and group.sorty_times or '<curr_group is None>',
                self.groups._maxes,
            )
        )
        self.debug('\n{}'.format(self.debug__str))

    def curr_group_add(self, some_fact):
        # The new fact is not yet wired.
        self.controller.affirm(some_fact.next_fact is None)
        self.controller.affirm(some_fact.prev_fact is None)
        # The new fact is not a known fact. (Because of groups'
        # time_since/time_until windows, we shouldn't find Facts
        # from the store amongst open time whose PKs we've seen.)
        self.controller.affirm(some_fact.pk not in self.by_pk.keys())
        self.new_fact_wire_links(some_fact)

        with self.fact_group_rekeyed():
            self.curr_group.add(some_fact)
            self.by_pk[some_fact.pk] = some_fact

    def new_fact_wire_links(self, some_fact):
        # 2019-02-13: (lb): Just a *momentaneous* FYI. (Feature should be all wired now.)
        if some_fact.start == some_fact.end:
            self.controller.client_logger.warning(
                'Found MOMENTANEOUS: {}'.format(some_fact.short),
            )

        if self.curr_fact.start == some_fact.end:
            self.controller.affirm(self.curr_fact.prev_fact is None)
            self.curr_fact.prev_fact = some_fact
            some_fact.next_fact = self.curr_fact
        if self.curr_fact.end == some_fact.start:
            self.controller.affirm(self.curr_fact.next_fact is None)
            self.curr_fact.next_fact = some_fact
            some_fact.prev_fact = self.curr_fact

    # ***

    @property
    def facts(self):
        for group in self.groups:
            for fact in group.facts:
                yield fact

    # ***

    def is_final_fact(self, edit_fact):
        group, index = self.locate_fact(edit_fact)
        if not group.until_time_stops:
            return False
        if not index == len(group) - 1:
            return False
        return True

    # ***

    def pop_final_gap_fact(self):
        final_fact = self[-1]
        if 'interval-gap' not in final_fact.dirty_reasons:
            return None
        group_fact = self.groups[-1].pop(-1)
        self.controller.affirm(group_fact is final_fact)
        return group_fact




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

from nark.items.activity import Activity

from ..crud.fact_dressed import FactDressed

__all__ = (
    'FactsManager_Gap',
)


class FactsManager_Gap(object):
    """"""

    def fact_from_interval_gap(self, since_time, until_time):
        self.controller.affirm((not until_time) or (since_time < until_time))
        self.last_fact_pk -= 1
        activity = Activity(name='')  # (lb): Meh. Seems like Fact should do.
        gap_fact = FactDressed(
            pk=self.last_fact_pk,
            activity=activity,
            start=since_time,
            end=until_time,
        )
        gap_fact.dirty_reasons.add('interval-gap')
        # Mark deleted until edited, so gap is not saved unless edited.
        gap_fact.deleted = True
        # No exceptions! All Fact copies must eventually lead to the original.
        gap_fact.orig_fact = gap_fact.copy()
        # Add to undo stack. Sorta tricky. Sorta a hack.
        self.on_insert_fact(gap_fact)
        return gap_fact

    def wire_two_facts_neighborly(self, fact_1, fact_2):
        self.controller.affirm(fact_1 < fact_2)
        self.controller.affirm(fact_2.prev_fact is None)
        self.controller.affirm(fact_1.next_fact is None)
        fact_1.next_fact = fact_2
        fact_2.prev_fact = fact_1




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

# MAYBE/2019-12-05: (lb): Decouple dob-viewer from click, used only for term. size.
import click_hotoffthehamster as click  # merely for get_terminal_size.

from nark.helpers.format_text import format_value_truncate
from nark.helpers.objects import resolve_attr_or_method

from dob_bright.termio import attr, fg

__all__ = (
    'FactsDiff',
)


# MAYBE: (lb): Move this module (FactsDiff) to dob?
#   And remove Fact.friendly_diff(), too?
# Because it's only used by the CLI...


class FactsDiff(object):
    """"""
    def __init__(self, orig_fact, edit_fact, formatted=False):
        self.orig_fact = orig_fact
        self.edit_fact = edit_fact
        self.formatted = formatted
        self.include_newlines = False
        self.exclude_attrs = None

    def __str__(self):
        return 'FactsDiff:\n- orig: {}\n- edit: {}'.format(
            self.orig_fact.short, self.edit_fact.short,
        )

    # ***

    def friendly_diff(
        self,
        truncate=False,
        exclude=None,
        show_elapsed=False,
        show_midpoint=False,
        show_now=False,
    ):
        def _friendly_diff():
            self.include_newlines = True
            self.exclude_attrs = exclude

            result = '' if not self.formatted else []
            result = assemble_diff_attrs(result)

            self.include_newlines = False
            self.exclude_attrs = None

            if not self.formatted:
                result = result.rstrip()
            else:
                while (len(result) > 0) and (not result[-1][1].strip()):
                    result.pop()

            return result

        def assemble_diff_attrs(result):
            result += self.diff_line_assemble(
                None, self.time_humanize(show_now), 'interval',
            )
            if show_midpoint:
                result += self.diff_line_assemble(
                    None, self.time_midpoint(), 'midpoint',
                )
            if show_elapsed:
                self_val, other_val = self.diff_time_elapsed(show_now)
                result += self.diff_line_assemble(
                    self_val, other_val, 'duration',
                )
            result += self.diff_attrs('start_fmt_local', 'start')
            if not show_now:
                result += self.diff_attrs('end_fmt_local', 'end')
            else:
                result += self.diff_attrs('end_fmt_local_nowwed', 'end')
            if (not truncate) or self.orig_fact.pk or self.edit_fact.pk:
                result += self.diff_attrs('pk', 'id', beautify=self.beautify_pk)
            result += self.diff_attrs('deleted', 'deleted')
            # MAYBE?: (lb): Would we even want to show the split_from fact?
            #  result += self.diff_attrs('split_from', 'split_from')
            result += self.diff_attrs('activity_name', 'activity')
            result += self.diff_attrs('category_name', 'category')
            if not self.formatted:
                result += self.diff_attrs('oid_tags', 'tags', colorful=True)
            else:
                result += self.diff_attrs('tags_tuples', 'tags', colorful=True)
            result += self.diff_attrs('description', 'description', truncate=truncate)
            return result

        # ***

        return _friendly_diff()

    # ***

    def diff_attrs(
        self,
        prop,
        name=None,
        style_class='',
        truncate=False,
        beautify=None,
        mouse_handler=None,
        **kwargs
    ):
        if (self.exclude_attrs is not None) and (name in self.exclude_attrs):
            return ''
        self_val = resolve_attr_or_method(self.orig_fact, prop, **kwargs)
        other_val = ''
        if self.edit_fact is not None:
            other_val = resolve_attr_or_method(self.edit_fact, prop, **kwargs)
            if callable(other_val):
                other_val = other_val()
            self_val, other_val = self.diff_values_enhance(
                self_val,
                other_val,
                style_class=style_class,
                truncate=truncate,
                beautify=beautify,
                mouse_handler=mouse_handler,
            )
        elif truncate:
            self_val = self.format_value_truncate(self_val)
            self_val = self.format_prepare(
                self_val, style_class=style_class, mouse_handler=mouse_handler,
            )
            other_val = self.format_prepare(other_val, style_class=style_class)
        attr_diff = self.diff_line_assemble(self_val, other_val, name)
        return attr_diff

    def diff_line_assemble(self, self_val, other_val, name=None):
        prefix = self.diff_values_padded_prefix(name)
        if not self.formatted:
            return self.diff_line_inline_style(self_val, other_val, prefix)
        else:
            return self.diff_line_tuples_style(self_val, other_val, prefix)

    def diff_values_enhance(
        self,
        self_val,
        other_val,
        style_class='',
        truncate=False,
        beautify=None,
        mouse_handler=None,
    ):
        differ = False
        if self_val != other_val:
            differ = True
        if truncate:
            self_val = self.format_value_truncate(self_val)
            other_val = self.format_value_truncate(other_val)
        if beautify is not None:
            self_val, other_val = beautify(self_val, other_val)
            if self_val != other_val:
                differ = True
        if differ:
            self_val = self.format_edited_before(self_val)
            self_val, other_val = self.format_edited_after(self_val, other_val)
        else:
            self_val = self.format_prepare(
                self_val, style_class=style_class, mouse_handler=mouse_handler,
            )
            other_val = self.format_prepare('', style_class=style_class)
        return (self_val, other_val)

    def format_prepare(self, some_val, style_class='', mouse_handler=None):
        if not self.formatted or not isinstance(some_val, str):
            # tags, e.g.,:
            #   [('fg: #C6C6C6 underline', '#'), ('fg: #D7FF87 underline', 'my-tag')]
            if (
                (mouse_handler is not None)
                and (isinstance(some_val, list))
                and (len(some_val) > 0)
                and (isinstance(some_val[0], tuple))
                and (len(some_val[0]) == 2)
            ):
                return [(tup[0], tup[1], mouse_handler,) for tup in some_val]
            return some_val
        if mouse_handler is None:
            return [(style_class, some_val)]
        return [(style_class, some_val, mouse_handler)]

    def format_value_truncate(self, val):
        # MAGIC_NUMBER: (lb): A third of the terminal (1 / 3.).
        # MAYBE/2019-02-15: Should have Carousel tells us width.
        term_width = click.get_terminal_size()[0]
        trunc_width = int(term_width * (1 / 3.))
        return format_value_truncate(val, trunc_width)

    # ***

    def diff_time_elapsed(self, show_now=False, style_class=''):
        self_val = self.time_elapsed(self.orig_fact, show_now)
        other_val = self.time_elapsed(self.edit_fact, show_now)
        if not self_val:
            # Make 'em the same, i.e., show no diff, no styling.
            self_val = other_val
        return self.diff_values_enhance(self_val, other_val, style_class=style_class)

    def time_elapsed(self, fact, show_now=False):
        # NOTE: start and/or end might be string; e.g., clock or rel. time.
        if (not fact.times_ok) and (not show_now):
            return None
        time_val = fact.format_delta(style='HHhMMm')
        return time_val

    def time_midpoint(self):
        return self.format_prepare(
            self.edit_fact.time_of_day_midpoint
        )

    def time_humanize(self, show_now=False):
        return self.format_prepare(
            self.edit_fact.time_of_day_humanize(show_now=show_now)
        )

    def beautify_pk(self, self_val, other_val):
        # (lb): NOTE: This is the only dirty_reasons usage in nark
        #               (most of its usage is in dob).
        if (
            'split' in self.edit_fact.dirty_reasons
            or 'split' in self.orig_fact.dirty_reasons
        ):
            pass
        if 'lsplit' in self.edit_fact.dirty_reasons:
            other_val = 'New split fact, created before new fact'
        if 'rsplit' in self.edit_fact.dirty_reasons:
            other_val = 'New split fact, created after new fact'
        return (self_val, other_val)

    # ***

    def format_edited_before(self, before_val):
        if not self.formatted:
            return '{}{}{}'.format(
                fg('spring_green_3a'),
                before_val,
                attr('reset'),
            )
        spring_green_3a = '00AF5F'
        style = 'fg:#{}'.format(spring_green_3a)
        before_parts = []
        if isinstance(before_val, str):
            before_parts += [(style, before_val)]
        elif before_val is not None:
            for tup in before_val:
                before_parts.append((style, tup[1]))
        return before_parts

    def format_edited_after(self, self_val, other_val):
        if not self.formatted:
            return '{}{}{}{}{} | was: '.format(
                attr('bold'),
                attr('underlined'),
                fg('light_salmon_3b'),
                other_val,
                attr('reset'),
                # (lb): What, emphasis has no italic option?
            ), self_val
        light_salmon_3b = 'D7875F'
        style = 'fg:#{} bold underline'.format(light_salmon_3b)
        after_parts = []
        if isinstance(other_val, str):
            after_parts += [(style, other_val)]
        elif other_val is not None:
            for tup in other_val:
                after_parts.append((style, tup[1]))
        # (lb): Swap the order, for display purposes.
        #   (These formatting functions are so janky!)
        if self_val and self_val[0][1]:
            after_parts += [('italic', ' | was: ')]
        return after_parts, self_val

    # ***

    def diff_values_padded_prefix(self, name):
        if name is None:
            return ''
        prefix_prefix = '  '
        padded_prefix = '{}{:.<19} : '.format(prefix_prefix, name)
        return padded_prefix

    def diff_line_inline_style(self, self_val, other_val, prefix=''):
        format_inline = '{}{}{}'.format(prefix, self_val or '', other_val or '')
        format_inline += "\n" if self.include_newlines else ''
        return format_inline

    def diff_line_tuples_style(self, self_val, other_val, prefix='', style_class=''):
        format_tuples = []
        if prefix:
            format_tuples += [(style_class, prefix)]
        if self_val:
            format_tuples += self_val
        if other_val:
            format_tuples += other_val
        if self.include_newlines:
            format_tuples += [('', '\n')]
        return format_tuples




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""
Fixtures available to the tests/.

- In general, fixtures should return a single instance.

- If a fixture is a factory, its name should reflect that.

- A fixture that is parametrized should be suffixed with
  ``_parametrized`` to imply it has increased complexity.
"""

import pytest

# For the sourced fixtures to work, either import specifically, or glob it all.
# noqa: F401 'foo.bar' imported but unused
# noqa: F403 'from foo.bar import *' used; unable to detect undefined names
from dob_bright.tests.conftest import *  # noqa: F401, F403

from dob_viewer.crud.fact_dressed import FactDressed


@pytest.fixture
def test_fact_cls():
    return FactDressed




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""~/.config/dob/styling/stylit.conf definition and encapsulating class."""

from gettext import gettext as _

from prompt_toolkit.layout.containers import Container
from prompt_toolkit.widgets.base import Label
from prompt_toolkit.widgets.base import TextArea

# MAYBE/2019-12-02: (lb) Is using normal stdout to print errors okay?
#                        Or should we use the Carousel (e.g., PPT modal)?
from dob_bright.termio import dob_in_user_warning

from .styling_stylit import create_stylit_object

__all__ = (
    'StylingConfig',
)


class StylingConfig(object):
    """Encapsulate Carousel-specific user configurable styling, PPT class names,
    and Pygments style syntax."""

    def __init__(self, matches_style):
        # The caller passes a dict-like object of rules read from
        # the user's stylit.conf file. We convert that to a StylitRoot
        # object, which encapsulates business logic in a dict-like
        # candy wrapper.
        self.rulesets = self.consume_stylit_conf(matches_style)

        # As the Carousel builds the PPT UX and creates and add components
        # to it, it'll register each stylable component. At that time, we
        # will find all the applicable user-defined rules (from self.rulesets),
        # and we'll cache them here, in self.componentry. (We'll also see what
        # rules apply to the component, but if the rules are applied a second
        # time, at least the rulesets will be cached.)
        # - tl;dr, A ruleset cache for each of the stylable PPT component:
        self.componentry = {}

    # ***

    def consume_stylit_conf(self, matches_style):

        def _consume_stylit_conf():
            if matches_style is None:
                return {}
            return _create_rulesets(matches_style)

        def _create_rulesets(matches_style):
            rulesets = {}
            for section, rules in matches_style.items():
                ruleset = create_ruleset_from_rules(rulesets, rules)
                rulesets[section] = ruleset
            return rulesets

        def create_ruleset_from_rules(rulesets, rules):
            ruleset = create_stylit_object()
            unconsumed = ruleset.update_known(rules)
            warn_user_any_unconsumed(unconsumed)
            return ruleset

        def warn_user_any_unconsumed(unconsumed):
            if not unconsumed:
                return
            msg = _(
                "The stylit.conf contains unknown settings: {}"
            ).format(unconsumed)
            dob_in_user_warning(msg)

        return _consume_stylit_conf()

    # ***

    def add_stylable_classes(self, ppt_widget, friendly_name, fact):
        # Here's an example stylit.conf contents that'll get you here:
        #   $ cat ~/.config/dob/styling/stylit.conf
        #   [My Category Style]
        #   category_name = 'My Category'
        #   scrollable_frame = class:my-category
        # and then your corresponding class could be defined as:
        #   $ cat ~/.config/dob/styling/styles.conf
        #   [my-style]
        #   my-category = 'bg:#CA85AC #000000'
        # and then wire it all via the config:
        #   $ dob config set style my-style

        def _add_stylable_classes():
            rulesets = rules_for_component(friendly_name)
            return add_stylable_if_triggered(ppt_widget, rulesets, fact)

        # ***

        def rules_for_component(friendly_name):
            try:
                return self.componentry[friendly_name]
            except KeyError:
                return assemble_component_rulesets(friendly_name)

        def assemble_component_rulesets(friendly_name):
            rebuild_component_rulesets_list(friendly_name)
            return self.componentry[friendly_name]

        def rebuild_component_rulesets_list(friendly_name):
            self.componentry[friendly_name] = {}
            for section, ruleset in self.rulesets.items():
                # The friendly_name is the name of the component in the UX.
                # The user uses the friendly_name in the stylit.conf to add
                # a class string to the component. Check here if it's empty,
                # meaning the user does not have this setting in their conf
                # (and the default '' was used); or, the user specified the
                # setting by set it to the empty string.
                if not ruleset[friendly_name]:
                    continue
                if ruleset['disabled']:
                    continue
                # We're just building a convenience lookup of rules that apply
                # to the named component. We'll run the rules check later.
                self.componentry[friendly_name][section] = ruleset

        # ***

        def add_stylable_if_triggered(ppt_widget, rulesets, fact):
            accumulated = ''
            for section, ruleset in rulesets.items():
                if not ruleset_triggered(section, ruleset, fact):
                    continue
                accumulated += apply_stylable_class(ppt_widget, ruleset)
            return accumulated

        # ***

        def ruleset_triggered(section, ruleset, fact):
            # NOTE: If more than one rules apply, we assume AND (because
            #       user can OR simply by using additional [ruleset]s).
            triggered = False

            # This is a tad bit funky: returning False because not triggered
            #   could be because the user did not configure any conditionals for
            #   this ppt_widget, or because one of the conditionals failed.
            # There are two False paths:
            #   - We want to return False either *after* checking that there are no
            #     conditionals, or;
            #   - We want to return False on the first conditional to fail;
            # As such, we use a trinary, aka three-level digital logic, -1, 0, 1.
            #   - 0 means the probe did not find any conditionals to test.
            #   - 1 means there were one or more conditionals, and all passed.
            #   - -1 means there were one or more conditionals, and one failed.
            # So on 0 or 1, keep processing, but on -1 or all 0s, return False;
            # and only return True if all 1s.

            trinary = probe_fact(section, ruleset, fact)
            if trinary == -1:
                return False
            triggered = trinary == 1 or triggered

            trinary = probe_eval(section, ruleset, fact)
            if trinary == -1:
                return False
            triggered = trinary == 1 or triggered

            return triggered

        def probe_fact(section, ruleset, fact):
            # For each attribute, there are multiple rules the user can choose
            # from to specify one of more names, e.g., one or more Activity
            # names might share the same style, and user could specify:
            #   [my-stylit-section]
            #   activities = Name1, Name2
            # or, if the user wants to be absurd, they could instead specify:
            #   [my-stylit-section]
            #   activity = Name1
            #   activities = Name2
            # because we'll just combine both rules and check for one to match.
            # - But for tags, the user might want to distinguish between
            #   AND and OR, so we'll go to the trouble to let the user
            #   specify if we should AND tags, or OR them.

            trinary = 0

            activities = set()
            if ruleset['activity']:
                activities.add(ruleset['activity'])
            if ruleset['activities']:
                activities = activities.union(ruleset['activities'])
            if activities:
                if fact.activity_name in activities:
                    trinary = 1
                else:
                    return -1

            categories = set()
            if ruleset['category']:
                categories.add(ruleset['category'])
            if ruleset['categories']:
                categories = categories.union(ruleset['categories'])
            if categories:
                if fact.category_name in categories:
                    trinary = 1
                else:
                    return -1

            # (lb): Because I'm too accommodating, a zillion ways to conditional tags.

            tags_any = set()
            if ruleset['tag']:
                tags_any.add(ruleset['tag'])
            if ruleset['tags']:
                tags_any = tags_any.union(ruleset['tags'])
            if ruleset['tags-any']:
                tags_any = tags_any.union(ruleset['tags-any'])
            if ruleset['tags-or']:
                tags_any = tags_any.union(ruleset['tags-or'])

            tags_all = set()
            if ruleset['tags-all']:
                tags_all = tags_all.union(ruleset['tags-all'])
            if ruleset['tags-and']:
                tags_all = tags_all.union(ruleset['tags-and'])

            if tags_any or tags_all:
                fact_tags = set([tag.name for tag in fact.tags])
                if tags_any:
                    if tags_any.intersection(fact_tags):
                        trinary = 1
                    else:
                        return -1
                if tags_all:
                    if not tags_all.difference(fact_tags):
                        trinary = 1
                    else:
                        return -1

            return trinary

        def probe_eval(section, ruleset, fact):
            # (lb): `'<str>' in ruleset` sends 0, not '<str>', to __getitem__?!
            # So do `'__eval__' not in ruleset.keys()`. Except no need, because
            # the key should exist, because create_stylit_object.
            compiled_code = ruleset['__eval__']
            if compiled_code is None:
                return 0
            try:
                trinary = eval(compiled_code)
            except Exception as err:
                msg = _(
                    "eval() failed on ‘eval’ stylit rule from “{0}”: {1}"
                ).format(section, str(err))
                # MAYBE/2019-12-02: (lb): Show errors in Carousel?
                # - Also one of few places where traverser imports ...helpers
                #   (and I want to make traverser less dob-dependent (coupled)).
                dob_in_user_warning(msg)
                # Such that we never do this error dance again!
                ruleset['__eval__'] = None
                return False
            return trinary and 1 or -1

        # ***

        def apply_stylable_class(ppt_widget, ruleset):
            custom_classes = ' {}'.format(ruleset[friendly_name])
            if ppt_widget is None:
                # Style being used in a (style, text, handler) tuple.
                pass
            elif isinstance(ppt_widget, Label):
                apply_stylable_to_label(ppt_widget, custom_classes)
            elif isinstance(ppt_widget, Container):
                apply_stylable_to_container(ppt_widget, custom_classes)
            elif isinstance(ppt_widget, TextArea):
                apply_stylable_to_text_area(ppt_widget, custom_classes)
            else:
                # Unexpected path. Unhandled PPT type.
                # (lb): This is a wee bit coupled to the whole application
                # framework: Get the Context from Click, and from that, get
                # our Controller object, which has an affirm method (which
                # is wired to the dob config's catch_errors).
                # MAYBE/2019-12-05: (lb): Decouple dob-viewer from click,
                #   used only for term. size.
                #   - Except here, where it's used to call our affirm, but
                #     also honors user's dev.catch_errors config setting.
                #     - Ideally, dob would set this up when it calls Carousel;
                #       or, dob, could configure dob-bright, and maybe dob-bright
                #       could offer an affirm() singleton or module method
                #       that is enabled after the config is read.
                import click_hotoffthehamster as click
                click.get_current_context().obj.affirm(False)
                pass
            return custom_classes

        def apply_stylable_to_label(label, custom_classes):
            # (lb): I'm totally wingin' it, in the sense that this works,
            # but I'm sure I'm bending the rules. The Label object has
            # three children we can play with to affect style:
            #   ∗ label.text, a list of (style, text, mouse-handler) tuples;
            #   ∗ label.formatted_text_control, which contains the text; and
            #   ∗ label.window, which extends further right to the parent edge.
            # Either of the first two options sets the style of just the text;
            # setting the window option affects the style to the parent edge.
            # - Because to be complicated, we let the user customize either
            # just the text, or the text and line padding.
            #   - We use the magical suffix, "-line", to decide which it is, e.g.,
            #         value-activity = class:my-style-text-only
            #         value-activity-line = class:my-style-whole-line
            if not friendly_name.endswith('-line'):
                # If label.text is tuples, their style beats formatted_text_control,
                # so rebuild the tuples list if present.
                if (
                    (isinstance(label.text, list))
                    and (len(label.text) > 0)
                    and (isinstance(label.text[0], tuple))
                    and (len(label.text[0]) > 1)
                ):
                    # Discard tup[0] (style component) and replace.
                    label.text = [
                        (custom_classes, tup[1], *tup[2:]) for tup in label.text
                    ]
                else:
                    # Otherwise, label.text is a simple 'string' (not a tuple),
                    # so we can use the formatted_text_control style.
                    label.formatted_text_control.style += custom_classes
            else:
                label.window.style += custom_classes

        def apply_stylable_to_container(container, custom_classes):
            container.style += custom_classes

        def apply_stylable_to_text_area(container, custom_classes):
            container.window.style += custom_classes

        # ***

        return _add_stylable_classes()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""~/.config/dob/styling/stylit.conf definition and encapsulating class."""

from gettext import gettext as _

from config_decorator import section

__all__ = (
    'create_stylit_object',
)


def create_stylit_object():

    COMPONENTRY_CLASSIFIER_HELP = _(
        "For component.style or (style, text) tuple, e.g., 'class:my-class1 fg:#000000'."
    )

    @section(None)
    class StylitRoot(object):

        def __init__(self):
            pass

    # ***

    @StylitRoot.section(None)
    class StylitRuleset(object):
        """"""

        def __init__(self):
            pass

        # ***

        @property
        @StylitRoot.setting(
            _("If True, skip this ruleset."),
        )
        def disabled(self):
            return False

        # ***

        @property
        @StylitRoot.setting(
            _("ADVANCED: Optional code to run in context of rules evaluator."),
        )
        def eval(self):
            # E.g.,
            #   eval = fact.category_name == 'My Category'
            return ''

        @property
        @StylitRoot.setting(
            _("Generated value."),
            name='__eval__',
            hidden=True,
            # Not necessary, because we generate the value, but could say:
            #   validate=inspect.iscode,
        )
        def compiled_eval(self):
            return None

        # ***

        @property
        @StylitRoot.setting(
            _("Match Facts with the specified Activity name."),
        )
        def activity(self):
            return ''

        # ---

        @property
        @StylitRoot.setting(
            _("Match Facts with specified comma-separated list of Activities."),
        )
        def activities(self):
            return []

        # ***

        @property
        @StylitRoot.setting(
            _("Match Facts with the specified Category name."),
        )
        def category(self):
            return ''

        # ---

        @property
        @StylitRoot.setting(
            _("Match Facts with specified comma-separated list of Categories."),
        )
        def categories(self):
            return []

        # ***

        @property
        @StylitRoot.setting(
            _("Match Facts with the specified Tag."),
        )
        def tag(self):
            return ''

        # ---

        @property
        @StylitRoot.setting(
            _("Match Facts with any of the matching tags."),
        )
        def tags(self):
            return []

        # ...

        @property
        @StylitRoot.setting(
            _("Match Facts with any of the matching tags."),
            name='tags-any',
        )
        def tags_any(self):
            return []

        # (lb): I'm indecisive. tags-any, or tags-or, or both??
        # - We can just 'hidden' one of them, and still let users decide.
        @property
        @StylitRoot.setting(
            _("Match Facts with any of the matching tags."),
            name='tags-or',
            hidden=True,
        )
        def tags_or(self):
            return []

        # ...

        @property
        @StylitRoot.setting(
            _("Match Facts with *all* of the matching tags."),
            name='tags-all',
        )
        def tags_all(self):
            return []

        # (lb): I'm indecisive. tags-all, or tags-and, or both??
        # - We can just 'hidden' one of them, and still let users decide.
        @property
        @StylitRoot.setting(
            _("Match Facts with *all* of the matching tags."),
            name='tags-and',
            hidden=True,
        )
        def tags_and(self):
            return []

        # ***

        # (lb): See Backlog for Feature Request to add additional conditionals,
        #   e.g., for start/end/duration.
        # Until then (if ever), use the 'eval' conditional to work with time
        #   or any of the other Fact attributes that are not magically wired
        #   to a Stylit setting.

    # ***

    @StylitRoot.section(None)
    class StylitClassify(object):
        """"""

        def __init__(self):
            pass

        # ***
        # ***

        # (lb): I feel like I should DRY up all the very similar method
        # definitions below (and, e.g., make a generator to create them).
        #
        # - On the other hand, having the long-winded method definitions
        #   below makes it pretty clear what's happening, and clearly
        #   shows what are the acceptable config keys and values.

        # ***
        # *** APPLICATION STREAMER AND ADJACENT EMPTY LINES
        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
        )
        def streamer(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='streamer-line',
        )
        def streamer_line(self):
            return ''

        # ***
        # *** HEADER META LINES -- TITLES HALF (LEFT SIDE)
        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-normal',
        )
        def title_normal(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-normal-line',
        )
        def title_normal_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-focus',
        )
        def title_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-focus-line',
        )
        def title_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-duration',
        )
        def title_duration(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-duration-line',
        )
        def title_duration_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-start',
        )
        def title_start(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-start-line',
        )
        def title_start_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-start-focus',
        )
        def title_start_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-start-focus-line',
        )
        def title_start_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-end',
        )
        def title_end(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-end-line',
        )
        def title_end_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-end-focus',
        )
        def title_end_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-end-focus-line',
        )
        def title_end_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-activity',
        )
        def title_activity(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-activity-line',
        )
        def title_activity_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-category',
        )
        def title_category(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-category-line',
        )
        def title_category_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-tags',
        )
        def title_tags(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='title-tags-line',
        )
        def title_tags_line(self):
            return ''

        # ***
        # *** HEADER META LINES -- VALUES HALF (RIGHT SIDE)
        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-normal',
        )
        def value_normal(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-normal-line',
        )
        def value_normal_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-focus',
        )
        def value_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-focus-line',
        )
        def value_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-duration',
        )
        def value_duration(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-duration-line',
        )
        def value_duration_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-start',
        )
        def value_start(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-start-line',
        )
        def value_start_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-start-focus',
        )
        def value_start_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-start-focus-line',
        )
        def value_start_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-end',
        )
        def value_end(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-end-line',
        )
        def value_end_line(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-end-focus',
        )
        def value_end_focus(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-end-focus-line',
        )
        def value_end_focus_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-activity',
        )
        def value_activity(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-activity-line',
        )
        def value_activity_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-category',
        )
        def value_category(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-category-line',
        )
        def value_category_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-tags',
        )
        def value_tags(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='value-tags-line',
        )
        def value_tags_line(self):
            return ''

        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='blank-line',
        )
        def blank_line(self):
            return ''

        # ***
        # *** CONTENT AREA CONDITIONAL STYLIT
        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='content-fact',
        )
        def scrollable_frame(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='content-help',
        )
        def content_help(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='interval-gap',
        )
        def interval_gap(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='unsaved-fact',
        )
        def unsaved_fact(self):
            return ''

        # ***
        # *** FACT ID CONDITIONAL STYLIT (LOWER LEFT CORNER)
        # ***

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='footer',
        )
        def footer_normal(self):
            return ''

        @property
        @StylitRoot.setting(
            COMPONENTRY_CLASSIFIER_HELP,
            name='footer-fact-id',
        )
        def footer_fact_id(self):
            return ''

        # ***
        # ***

    return StylitRoot




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact Editing State Machine"""

from collections import namedtuple

from nark.items.fact import Fact

from dob_bright.termio.style import colorize, set_coloring

from ..ptkui.ppt_markup import namilize

from .facts_diff import FactsDiff

__all__ = (
    'FactDressed',
    # PRIVATE:
    #  'FactoidSource',
)


FactoidSource = namedtuple(
    'FactoidSource', ('line_num', 'line_raw'),
)


class FactDressed(Fact):
    """"""

    def __init__(
        self,
        *args,
        dirty_reasons=None,
        line_num=None,
        line_raw=None,
        **kwargs
    ):
        super(FactDressed, self).__init__(*args, **kwargs)
        # For tracking edits between store saves.
        self.dirty_reasons = dirty_reasons or set()
        # For identifying errors in the input.
        self.parsed_source = FactoidSource(line_num, line_raw)
        self.orig_fact = None
        # For Carousel (dob-viewer) navigation.
        self.next_fact = None
        self.prev_fact = None

    # ***

    @property
    def short(self):
        friendly = (
            '0x{:12x} / 🏭 {} / {} to {:23} / prev: {:12x} / next: {:12x}'.format(
                id(self),
                self.pk is not None and '{:6d}'.format(self.pk) or '<None>',
                self.start_fmt_local,
                self.end_fmt_local or '..........now..........',
                self.prev_fact and id(self.prev_fact) or 0,
                self.next_fact and id(self.next_fact) or 0,
            )
        )
        return friendly

    # ***

    def copy(self, *args, **kwargs):
        """
        """
        new_fact = super(FactDressed, self).copy(*args, **kwargs)
        new_fact.dirty_reasons = set(list(self.dirty_reasons))
        new_fact.parsed_source = self.parsed_source
        new_fact.orig_fact = self.orig_fact or self
        # SKIP: next_fact, prev_fact.
        return new_fact

    # ***

    def friendly_diff(self, other, formatted=False, **kwargs):
        facts_diff = FactsDiff(self, other, formatted=formatted)
        return facts_diff.friendly_diff(**kwargs)

    # ***

    def friendly_str(self, *args, colorful=False, **kwargs):
        was_coloring = set_coloring(colorful)
        friendly_str = super(FactDressed, self).friendly_str(*args, **kwargs)
        set_coloring(was_coloring)
        return friendly_str

    def oid_colorize(self, oid_part, oid_text):
        """Stylizes parts of the Factoid with color and emphasis.
        """
        # FIXME: Turn this into config. (And check BonW vs WonB.)
        lookup = {
            'pk': ('grey_78'),
            'act@gory': ('cornflower_blue', 'bold', 'underlined'),
            '#': ('grey_78'),
            'tag': ('dark_olive_green_1b'),
            '#tag': ('underlined'),
            'start': ('sandy_brown'),
            'end': ('sandy_brown'),
            'to': ('grey_85'),
            'duration': ('grey_78'),
        }
        colorized = colorize(oid_text, *lookup[oid_part])
        return colorized

    # ***

    def squash(self, other, squash_sep=''):
        def _squash():
            # (lb): The squash is a useful end user application feature for existing
            # facts, and I'm not sure what else it might be used for, so I'm putting
            # a bunch of asserts here to force you to re-read this comment when next
            # this code blows up because new usage and you realize you can assuredly
            # delete this comment and one or all of these assert and you will likely
            # be just fine.
            assert other.pk is None or other.pk < 0
            assert not self.deleted
            assert not other.deleted
            assert not other.split_from
            # When squashing, the first fact should have a start, but not an end.
            # And we do not care about other; it could have a start, or an end, or
            # neither.
            assert self.start
            assert not self.end

            self.end = other.start or other.end

            if other.activity_name or other.category_name:
                # (lb): MAYBE: Do we care that this is destructive?
                self.activity = other.activity

            self.tags_replace(self.tags + other.tags)

            description_squash(other, squash_sep)

            self.dirty_reasons.add('squash')
            if self.end:
                self.dirty_reasons.add('stopped')
                self.dirty_reasons.add('end')

            other.deleted = True
            # For completeness, and to make verification easier.
            other.start = self.start
            other.end = self.end

            other.dirty_reasons.add('deleted-squashed')

        def description_squash(other, squash_sep=''):
            if not other.description:
                return
            # (lb): Build local desc. copy, because setter stores None, never ''.
            new_description = self.description or ''
            new_description += squash_sep if new_description else ''
            new_description += other.description
            self.description = new_description
            other.description = None

        _squash()

    # ***

    @classmethod
    def create_from_factoid(cls, factoid, *args, **kwargs):
        """Creates a new Fact from Factoid text, and sets bulk import metadata.
        """
        new_fact, err = super(FactDressed, cls).create_from_factoid(
            factoid, *args, **kwargs
        )
        if new_fact is not None:
            line_num = 1
            line_raw = factoid
            new_fact.parsed_source = FactoidSource(line_num, line_raw)
        return new_fact, err

    # ***

    @property
    def dirty(self):  # MAYBE: Rename: positive()?
        # MAYBE/FIXME: Set dirty_reasons if fact.pk < 0, on new FactDressed.
        return (
            (self.unstored or len(self.dirty_reasons) > 0)
            and ('interval-gap' not in self.dirty_reasons)
        )

    # *** Linked list methods.

    @property
    def has_next_fact(self):
        return self.next_fact is not None

    @property
    def has_prev_fact(self):
        return self.prev_fact is not None

    # *** Presentation concerns.

    def oid_tags(self, *args, colorful=False, **kwargs):
        was_coloring = set_coloring(colorful)
        friendly_tags = super(FactDressed, self).oid_tags(*args, **kwargs)
        set_coloring(was_coloring)
        return friendly_tags

    # NOTE/2020-01-27: (lb): Returns PPT tuple, not something dob knows about.
    # - We could move this method to dob-viewer, to truly decouple.
    #   But then dob-viewer needs to override FactDressed (self.store.fact_cls).
    #   (For now, I'm happy this method at least made it out of nark!)
    def tags_tuples(
        self,
        hashtag_token='#',
        quote_tokens=False,
        colorful=False,
        underlined=False,
        split_lines=False,
    ):

        def format_tagname(tag):
            # FIXME/2020-01-28: (lb): make underline and colors configable.
            underline_fmt = ' underline' if colorful else ''
            tagged = []
            #
            tclss_fmt = ' class:tag-{}'.format(namilize(tag.name))
            #
            token_fmt = 'fg: ' if (colorful or underline_fmt) else ''
            token_fmt += '#C6C6C6' if colorful else ''
            token_fmt += underline_fmt
            token_fmt += tclss_fmt
            tagged.append((token_fmt, hashtag_token))
            #
            tname_fmt = 'fg: ' if (colorful or underline_fmt) else ''
            tname_fmt += '#D7FF87' if colorful else ''
            tname_fmt += underline_fmt
            tname_fmt += tclss_fmt
            tagged.append((tname_fmt, tag.name))
            #
            if quote_tokens:
                fmt_quote = ('', '"')
                tagged.insert(0, fmt_quote)
                tagged.append(fmt_quote)
            return tagged

        # NOTE: The returned string includes leading space if nonempty!
        tagnames = []
        if self.tags:
            # Build array of PPT tuples.
            fmt_sep = ('', "\n") if split_lines else ('', ' ')
            n_tag = 0
            for fmtd_tagn in self.tagnames_sorted_formatted(format_tagname):
                if n_tag > 0:
                    tagnames += [fmt_sep]
                n_tag += 1
                tagnames += fmtd_tagn
        return tagnames

    # ***

    @property
    def html_notif(self):
        """
        A briefer Fact one-liner using HTML. Useful for, e.g., notifier toast.
        """
        was_coloring = set_coloring(False)
        duration = '[{}]'.format(self.format_delta(style=''))
        actegory = self.oid_actegory(omit_empty_actegory=True)
        actegory = actegory or '<i>No activity</i>'
        description = self.oid_description(cut_width=39, sep=': ')
        simple_str = (
            '{} {}{}'
            .format(
                duration,
                actegory,
                description,
            )
        )
        set_coloring(was_coloring)
        return simple_str

    # ***




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel Custom User (Pygment-defined) Styles."""

from gettext import gettext as _

# MEH/2019-12-05: (lb): We could decouple dob-viewer from click, which is only
# used for getting the terminal size. But dob requires Click, so it's there.
import click_hotoffthehamster as click  # merely for get_terminal_size.

from config_decorator import section

from ..ptkui.ppt_markup import namilize

__all__ = (
    'color',
    'default',
    'light',
    'night',
    # PRIVATE:
    # '_create_style_object',
    # '_stylize_all_one',
)


def _create_style_object():
    """Define and return a ConfigDecorator for ~/.config/dob/styling/styles.conf."""

    # (lb): We want to wrap the @setting decorator (aka decorate the decorator),
    # which means we need to use two classes -- one to define the decorator, and
    # another that can call the decorator after the first class has been defined.

    @section(None)
    class StylesRoot(object):
        """"""

        def __init__(self):
            pass

        # ***

        NOT_STYLE_CLASSES = set()

        @classmethod
        def setting_wrap(cls, *args, not_a_style=False, **kwargs):
            def decorator(func):
                name = kwargs.get('name', func.__name__)
                if not_a_style:
                    cls.NOT_STYLE_CLASSES.add(name)
                return StylesRoot.setting(*args, **kwargs)(func)
            return decorator

        @classmethod
        def collect_tups(cls):
            style_classes = [
                (namilize(key), val.value) for key, val in StylesRoot._key_vals.items()
                if key not in cls.NOT_STYLE_CLASSES
            ]
            return style_classes

    # Now that the @section re-decorator is defined, we declare a class that
    # uses it to build a settings config. Note that we use a little @section
    # decorator magic on this second class -- by using a None section name,
    # in @StylesRoot.section(None), it causes the class being decorated (in
    # this case, CustomScreen) to be an alias to the root config (StylesRoot)!
    # So the settings defined in CustomScreen will be added to the root config,
    # and not to a sub-section. Somewhat wonky, but wonky is our jam.

    # (lb): 2 hacks, because eccentric magic of @section and ConfigDecorator:
    #
    # 1.) We wrap @ConfigDecorator.setting so we can identify which of our
    #     key-value settings are "class:..." assignments, and which of our
    #     key-value settings are the rules to decide if the class assignments
    #     should be applied.
    #     - The reason we mix the two types of settings in the same config
    #     section, rather than using two separate sections, is to keep the
    #     config flat and simple. It helps minimize the types of errors the
    #     user can make while editing their stylit.conf file.
    #     - The hack is reaching into StylesRoot (which is also a hack: it's
    #     not a class, but an object instance! because of the eccentric magic
    #     of @section) and finding our @setting_wrap through the special
    #     _innerobj attribute.
    #
    # 2.) Hack number two here is specifying a `None` @section when defining
    #     the CustomScreen class, telling the @setting decorator to attach
    #     each setting to the root config, and not to a sub-section.
    #     - I.e., each setting herein will be applied to the StylesRoot object.
    #     In fact, CustomScreen disappears, in a sense, because the decorator
    #     returns the section to which the class settings apply, so you'll find:
    #       assert(CustomScreen is StylesRoot)  # Totally crazy, I know!
    #
    # 3?) I suppose there's actually a third hack, or maybe it's a trick, or just
    #     The Way To Do It: we use the encompassing _create_style_object() method
    #     to localize the defined classes/objects (StylesRoot and CustomScreen)
    #     so that we don't end up creating singletons, but rather create unique
    #     config objects each time.

    setting = StylesRoot._innerobj.setting_wrap

    def evaluate_content_dimension(value, dim):
        if isinstance(value, int):
            return value
        term_width, term_height = click.get_terminal_size()
        # E.g., = value.replace('term_width', str(term_width))
        evalable = value.replace(dim, str(locals()[dim]))
        compiled = compile(evalable, filename='<string>', mode='eval')
        executed = eval(compiled)
        return int(executed)

    @StylesRoot.section(None)
    class CustomScreen(object):
        """"""

        def __init__(self):
            pass

        # ***

        @property
        @setting(
            _("Generated value."),
            hidden=True,
            not_a_style=True,
        )
        def collect_tups(self):
            return StylesRoot._innerobj.collect_tups()

        # ***

        @property
        @setting(
            _("Name of the style to use for default values."),
            choices=['', 'default', 'night', 'light', 'color'],
            name='base-style',
            not_a_style=True,
        )
        def base_style(self):
            return ''

        # ***

        @property
        @setting(
            _("JUSTIFY/CENTER UX in terminal, or position LEFT or RIGHT."
              " See also: content-width."),
            choices=['LEFT', 'CENTER', 'RIGHT', 'JUSTIFY'],
            name='editor-align',
            not_a_style=True,
        )
        def editor_align(self):
            # (lb): AFAICT, 'CENTER' is equivalent to 'JUSTIFY'.
            return 'CENTER'

        # ***

        def evaluate_content_height(value):
            return evaluate_content_dimension(value, 'term_height')

        @property
        @setting(
            # NOTE: If content-height is set to None, the content height changes
            #       every time the user switches between Facts in the editor.
            _("Sizes content area height; may ref. term, e.g., “term_height - 15”."),
            name='content-height',
            not_a_style=True,
            # Use conform to change value used internally, but to keep the user's
            # input, e.g., when writing config to file, be sure to return, say,
            # "term_height - 15", and not the calculated internal value.
            conform=evaluate_content_height,
            allow_none=True,
        )
        def content_height(self):
            #
            # PPT will gripe "Window too small..." if we make the content
            # too large, and because we add a border box around the content,
            # the minimum height is:
            #
            #     minimum_height = term_height - 2
            #
            # except that accommodates *only* the content area, and
            # it pushes out (clips) the streamer, header, and footer.
            #
            # (lb): I'm hardcoding the height adjustment value (the 15, below),
            # in lieu of poking around our zone_details and other structures to
            # compute the height of each component. We'll just hardcode this now
            # and determine later if we should encode less UX knowledge here.
            #
            # Subtract from the terminal height the height of other components.
            # (I.e., grow the content area vertically to fill in the UX.)
            #
            # - The maths:
            #   - +1: There's 1 blank line above the streamer.
            #   - +3: The streamer is 3 lines, e.g.,
            #           ┌───────────────────────────────────────────────────────╮
            #           │ Wed 23 Jan 2019 ◐ 11:00 AM — 02:21 PM Thu 05 Dec 2019 │
            #           ╰───────────────────────────────────────────────────────╯
            #   - +1: There's 1 blank line between streamer and headers.
            #   - +6: +1 l. each: duration|start|end|activity|category|tags.
            #   - +1: There's 1 blank line between headers and content.
            #   - +2: The content area itself is bordered.
            #   - +1: The footer is a single line.
            #     --
            #     15

            # NOTE: There's no need to cache the calculated default value.
            # - (lb): This method is called only once per runtime, AFAICT.

            # We could `return click.get_terminal_size()[1] - 15` directly here,
            # but let's use a string and show off how to use the magic value. If
            # the user dumps the styling config, they'll see this string and have
            # a better idea how best to customize this value.

            # FIXME/2019-12-05: (lb): I bet this clips for Facts w/ 2+ lines of #tags.

            return 'term_height - 15'

        # ***

        def evaluate_content_width(value):
            return evaluate_content_dimension(value, 'term_width')

        @property
        @setting(
            # NOTE: To get 1 column of whitespace on both (the left and right)
            #       sides of the UX, subtract 3 (not 2) from the terminal width.
            #       E.g., in ~/.config/dob/styling/styles.conf:
            #         [my-style-1]
            #         content-width = term_width - 3
            # (lb): Here's nice dims that works well in many terminal sizings::
            #         [adaptive-style]
            #         content-width = 'min(term_width * 0.62432, term_height * 3.35988)'
            #         content-height = 'max(term_height - 22, 10)'
            _("Specifies UX width; may ref. curr. dims., e.g., “term_width - 3”."),
            name='content-width',
            not_a_style=True,
            # Use conform to change value used internally, but to keep the user's
            # input, e.g., when writing config to file, be sure to return, say,
            # "term_width - 3", and not the calculated internal value.
            conform=evaluate_content_width,
            allow_none=True,
        )
        def content_width(self):
            # We could return None and PPT would default to terminal width.
            # But let's use the magic string value, so when the user dumps
            # the styling config, they get a better idea what the default is.
            return 'term_width'

        # ***

        @property
        @setting(
            _("If True, wraps the content area text; otherwise, scrolls horizontally."),
            name='content-wrap',
            not_a_style=True,
        )
        def content_wrap(self):
            return True

        # ***

    # ***

    # The following class adds its members to the same root config object as the
    # previous class, but none of the settings below identify as a not_a_style.

    @StylesRoot.section(None)
    class CustomStyles(object):
        """"""

        # DRY alert: StylitClassify and CustomStyles have similarly-named methods,
        # but they're used differently.
        # - In StylitClassify, the methods are the (style) class names, and the
        #   values are the style definitions (e.g., bg and fg colors) that are
        #   used to make the PPT Style object. (The Style() object is used to
        #   lookup a class name and get the style definition).
        #   - Currently, the runtime makes the Style() object once, and never
        #     modifies it.
        # - In CustomStyles, the methods are also the (style) class names, but
        #   the values are class names and color styles (as a string) to add to
        #   the widget styles when a conditional rule applies.
        #   - I.e., the values in this class, when used by a conditional, are
        #     appended to the style of the widget that is identified by that
        #     class name. That is, the style string is not used to change the
        #     class name it references, but instead the class name refers to the
        #     widgets that identify with that class name. Got it?
        #     - E.g., if a conditional triggers and its styles are added to the
        #       activity value component, that component would have `style`:
        #         'class:value-normal class:value-activity class:{custom-value}'
        #       where {custom-value} is the value of the value-activity setting
        #       from this class (which gets the value from the user's stylit.conf).
        #       - Note that 'class:value-activity' (a string) is always added to
        #         the component, regardless of any conditionals, and its definition
        #         comes from the user's styles.conf.
        #       - By specifying 'value-activity' in stylit.conf, what ends up
        #         happening is that the conditional style is appended to the
        #         component style, following 'class:value-activity', thereby
        #         shadowing the basic style. What you have, in a sense, is
        #         'class:value-activity' serving as a default value, and stylit's
        #         'value-activity' being applied after that default.
        # (lb): I hope this isn't too confusing. But I think it'd be more confusing
        #   if the names were not the same in both configs. In any case, the methods
        #   listed here should be ordered like in the other class, to make it easier
        #   to keep the two config modules maintained and synced.
        # - Note that it's not necessary to define any of the attributes here,
        #   because the styles.conf parser treats any unknown setting from the
        #   user's conf as a class definition -- which has the same effect as
        #   declaring each @setting here. But by declaring them here, we can use
        #   the object for producing help, or to built a template file to make
        #   user onboarding easier.

        def __init__(self):
            pass

        # ***

        @property
        @setting(
            _("Styles the streamer UX banner (topmost UX)."),
        )
        def streamer(self):
            return ''

        @property
        @setting(
            _("Styles the streamer UX banner (topmost UX) including empty lines."),
            name='streamer-line',
        )
        def streamer_line(self):
            return ''

        # ***

        @property
        @setting(
            _("Styles the header titles."),
            name='title-normal',
        )
        def title_normal(self):
            return ''

        @property
        @setting(
            _("Styles the header titles, including adjacent whitespacing."),
            name='title-normal-line',
        )
        def title_normal_line(self):
            return ''

        @property
        @setting(
            _("Styles the header title whose value has focus and is editable."),
            name='title-focus',
        )
        def title_focus(self):
            return ''

        @property
        @setting(
            _("Styles the header title and its whitespacing, when its value focused."),
            name='title-focus-line',
        )
        def title_focus_line(self):
            return ''

        # ***

        # See below for other title-* settings: title-duration → title-tags-line.

        # ***

        @property
        @setting(
            _("Styles the header value text."),
            name='value-normal',
        )
        def value_normal(self):
            return ''

        @property
        @setting(
            _("Styles the header value line (text and the whitespace right of it)."),
            name='value-normal-line',
        )
        def value_normal_line(self):
            return ''

        @property
        @setting(
            _("Styles the header value text when it has focus and is editable."),
            name='value-focus',
        )
        def value_focus(self):
            return ''

        @property
        @setting(
            _("Styles the header value line of the editable value with focus."),
            name='value-focus-line',
        )
        def value_focus_line(self):
            return ''

        # ***

        # The remaining title-* settings: title-duration through title-tags-line.
        # And remaining value-* settings: value-duration through value-tags-line.

        for prefix in ('title', 'value'):
            for part in (
                'duration',
                'start',
                'start-focus',
                'end',
                'end-focus',
                'activity',
                'category',
                'tags',
            ):
                for suffix in ('', '-line'):
                    class_name = '{}-{}{}'.format(prefix, part, suffix)

                    is_line = suffix == '-line'
                    is_focus = part.endswith('-focus')
                    help_l = is_line and _(', including adjacent whitspace') or ''
                    help_f = is_focus and _(', when value focused') or ''
                    doc = _("{} {} style{}{}.").format(part, prefix, help_l, help_f)

                    @setting(doc, name=class_name)
                    def _title_setting(self):
                        return ''

        # ***

        @property
        @setting(
            _("Styles the empty line between the tags and the content area."),
            name='blank-line',
        )
        def blank_line(self):
            return ''

        # ***

        @property
        @setting(
            _("Style the content area when showing the Fact description."),
            name='content-fact',
        )
        def content_fact(self):
            return ''

        # ***

        @property
        @setting(
            _("Styles the content area when showing the one-page help."),
            name='content-help',
        )
        def content_help(self):
            return ''

        # ***

        @property
        @setting(
            _("Styles the content area when showing a generated, unsaved gap Fact."),
            name='interval-gap',
        )
        def interval_gap(self):
            return ''

        # ***

        @property
        @setting(
            _("Styles the content area when showing an edited, unsaved Fact."),
            name='unsaved-fact',
        )
        def unsaved_fact(self):
            return ''

        # ***

        @property
        @setting(
            _("Styles of footer section of the UX (bottommost line)."),
            name='footer',
        )
        def footer_normal(self):
            return ''

        @property
        @setting(
            _("Styles the Fact ID or Hot Notif text in the UX footer."),
            name='footer-fact-id',
        )
        def footer_fact_id(self):
            return ''

        # ***

    return StylesRoot


# ***

def default():
    """Default defines all options so tweaked stylings may omit any."""

    # (lb): Because of the magic of @section and how StylesRoot is not
    # really a class, but rather a ConfigDecorator object, we use a wrapper
    # function to both define the config (every time it's called) and to
    # return the newly instantiated ConfigDecorator object (which can only
    # be accessed using the name of the @section-decorated class!).

    styling = _create_style_object()

    return styling


# ***

def _stylize_all_one(styling, style):
    """"""

    styling['streamer'] = style
    # (lb): I cannot decide if I like the streamer bold or not.
    styling['streamer'] += ' bold'

    styling['streamer-line'] = style

    # Set all the title-* and value-* values,
    #   title-normal through value-tags-line.
    for prefix in ('title', 'value'):
        for part in (
            'normal',
            'focus',
            'duration',
            'start',
            'start-focus',
            'end',
            'end-focus',
            'activity',
            'category',
            'tags',
        ):
            for suffix in ('', '-line'):
                class_name = '{}-{}{}'.format(prefix, part, suffix)
                styling[class_name] = style

    styling['blank-line'] = style

    styling['content-fact'] = style
    styling['content-help'] = style
    styling['interval-gap'] = style
    styling['unsaved-fact'] = style

    styling['footer'] = style
    styling['footer-fact-id'] = style


# ***

def light():
    styling = default()
    _stylize_all_one(styling, 'bg:#FFFFFF #000000')
    return styling


# ***

def night():
    styling = default()
    _stylize_all_one(styling, 'bg:#000000 #FFFFFF')
    return styling


# ***

def color():
    styling = night()

    styling['streamer'] = 'fg:#5F5FFF bold'
    # styling['streamer-line'] = ...

    # styling['title-normal'] = ...
    #  through
    # styling['value-focus-line'] = ...

    # styling['blank-line'] = ...

    # Default Fact.description frame background.
    styling['content-fact'] = 'bg:#9BC2C2 #000000'
    # Fact.description background when showing help.
    styling['content-help'] = 'bg:#66AAAA #000000'
    # Other contextual Fact.description background colors.
    styling['interval-gap'] = 'bg:#AA6C39 #000000'
    styling['unsaved-fact'] = 'bg:#D0EB9A #000000'

    # styling['footer'] = ...
    # styling['footer-fact-id'] = ...

    return styling




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Fact Editing Start and End Time Adjuster"""

from datetime import timedelta

from .redo_undo_edit import UndoRedoTuple

__all__ = (
    'StartEndEdit',
)


class StartEndEdit(object):
    """"""
    def __init__(self, edits_manager):
        self.edits_manager = edits_manager
        self.controller = edits_manager.controller
        # MAYBE/2019-01-31: (lb): This class is tightly coupled.
        #  We might as well concede defeat and make this class a
        #  part of an EditsManager hierarchy (like the FactsManager
        #  classes).
        self.restore_facts = edits_manager.restore_facts
        self.editable_fact = edits_manager.editable_fact
        self.editable_fact_next = edits_manager.editable_fact_next
        self.editable_fact_prev = edits_manager.editable_fact_prev
        self.redo_undo = edits_manager.redo_undo

    # ***

    def edit_time_adjust(
        self,
        delta_mins_or_time,
        start_or_end,
        end_maybe=None,
        gap_okay=False,
    ):
        def _edit_time_adjust():
            edit_fact = self.editable_fact()
            edit_prev, edit_next = _edit_neighbors(edit_fact)
            newest_changes = _undoable_changes(edit_fact, edit_prev, edit_next)
            adjust_time(edit_fact, edit_prev, edit_next)
            adjust_time_fix_overlaps(edit_fact, edit_prev, edit_next)
            debug_log_facts('edit-time-final', edit_fact, edit_prev, edit_next)
            post_process_edited(newest_changes)

        # ***

        def _edit_neighbors(edit_fact):
            edit_prev = self.time_adjust_editable_prev(
                edit_fact, start_or_end, end_maybe,
            )
            edit_next = self.time_adjust_editable_next(
                edit_fact, start_or_end, end_maybe,
            )
            debug_log_facts('edit-time-begin', edit_fact, edit_prev, edit_next)
            return edit_prev, edit_next

        # ***

        def _undoable_changes(edit_fact, edit_prev, edit_next):
            try:
                context = delta_mins_or_time.total_seconds() >= 0 and 'pos' or 'neg'
            except AttributeError:
                context = start_or_end
            edit_what = 'adjust-time-{}'.format(context)
            # Get an UndoRedoTuple from a copy of the Facts we're about to edit.
            # And set UndoRedoTuple.altered to the new copies.
            newest_changes = self.redo_undo.undoable_changes(
                edit_what, edit_fact, edit_prev, edit_next,
            )
            return newest_changes

        # ***

        def adjust_time(edit_fact, edit_prev, edit_next):
            if isinstance(delta_mins_or_time, timedelta):
                adjust_time_by_delta(edit_fact, edit_prev, edit_next)
            else:
                adjust_time_set_time(edit_fact, edit_prev, edit_next)

        def adjust_time_by_delta(edit_fact, edit_prev, edit_next):
            adjust_time_by_delta_apply(edit_fact, edit_prev, 'start')
            adjust_time_by_delta_apply(edit_fact, edit_next, 'end')

        def adjust_time_by_delta_apply(edit_fact, prev_or_next, apply_start_or_end):
            # delta_mins_or_time is datetime.timedelta.
            self.edit_time_adjust_time(
                edit_fact,
                prev_or_next,
                delta_mins_or_time,
                apply_start_or_end,
                gap_okay,
                start_or_end,
                end_maybe,
            )

        def adjust_time_set_time(edit_fact, edit_prev, edit_next):
            # delta_mins_or_time is datetime.datetime.
            # Only one neighbor is not None, so don't try too hard.
            neighbor = edit_prev or edit_next
            self.edit_time_apply_time(
                edit_fact, delta_mins_or_time, neighbor, start_or_end, gap_okay,
            )

        def adjust_time_fix_overlaps(edit_fact, edit_prev, edit_next):
            # 2020-04-12: (lb): I haven't been in this code in a while, but
            # seems fishy we're setting neighbor's time here and not in
            # edit_time_apply_time.
            if edit_prev and edit_prev.end > edit_fact.start:
                self.controller.affirm(False)  # FIXME/2020-04-12: Use case?
                edit_prev.end = edit_fact.start
            if edit_next and edit_next.start < edit_fact.end:
                self.controller.affirm(False)  # FIXME/2020-04-12: Use case?
                edit_next.start = edit_fact.end

        # ***

        def post_process_edited(newest_changes):
            # (lb): Ug, more coupling. Because we are managing the redo-undo stack
            # specially here (possibly removing the latest undo, and squishing it
            # with the new edits), we do not want to call edits_manager.apply_edits,
            # which calls update_redo_undo_and_conjoined, which fiddles with redo_undo.
            # Instead, do things piecemeal: pop the latest undo; squish it; push it back;
            # clear the redo; and use the restore_facts method to fix wiring (update the
            # edits_manager.edit_facts and facts_manager.by_pk lookups, and update the
            # facts_manager fact-groups).

            # If same Facts edited with same tool within DISTINCT_CHANGES_THRESHOLD
            # time, pop the previous undo.
            last_undo_or_newest_changes = (
                self.redo_undo.remove_undo_if_same_facts_edited(
                    newest_changes,
                )
            )
            self.controller.affirm(
                (newest_changes is last_undo_or_newest_changes)
                or (newest_changes.pristine == last_undo_or_newest_changes.altered)
            )

            # Mark things dirty (or not).
            self.edits_manager.manage_edited_dirty_flags(newest_changes.altered)

            # Create the freshest undo. It might be squished with an earlier
            # undo created by the user within DISTINCT_CHANGES_THRESHOLD ago,
            # in which case, the pristine Facts are from the previous undo.
            # In all cases, the altered Facts are the ones we just edited.
            undoable = UndoRedoTuple(
                last_undo_or_newest_changes.pristine,
                newest_changes.altered,
                last_undo_or_newest_changes.time,
                last_undo_or_newest_changes.what,
            )

            # In lieu of having called add_undoable, add the changes to the undo stack.
            self.redo_undo.append_changes(
                self.redo_undo.undo,
                undoable,
                whence='edit_time_adjust',
            )
            # This invalidates the redo stack.
            self.redo_undo.clear_changes(self.redo_undo.redo, 'edit_time_adjust')

            self.restore_facts(
                newest_changes.altered,
                # This is a little tricky: the Facts currently wired in the
                # FactsManager were altered by the previous Undo that we might
                # be replacing if within DISTINCT_CHANGES_THRESHOLD, so use
                # these Facts to fix the wiring.
                newest_changes.pristine,
            )

        # ***

        def debug_log_facts(prefix, edit_fact, edit_prev, edit_next):
            self.controller.client_logger.debug(
                '{}\n- edit: {}\n- prev: {}\n- next: {}'.format(
                    prefix,
                    edit_fact and edit_fact.short or '<no such fact>',
                    edit_prev and edit_prev.short or '<no such fact>',
                    edit_next and edit_next.short or '<no such fact>',
                ),
            )

        # ***

        _edit_time_adjust()

    # ***

    def time_adjust_editable_prev(self, edit_fact, *attrs):
        if 'start' not in attrs:
            return None
        return self.editable_fact_prev(edit_fact)

    def time_adjust_editable_next(self, edit_fact, *attrs):
        if 'end' not in attrs:
            return None
        return self.editable_fact_next(edit_fact)

    # ***

    def edit_time_adjust_time(
        self, edit_fact, neighbor, delta_time, start_or_end, gap_okay, *attrs
    ):
        if start_or_end not in attrs:
            return

        curr_time = getattr(edit_fact, start_or_end)
        if curr_time is None:
            # The ongoing, un-ended, active Fact.
            self.controller.affirm(start_or_end == 'end')
            # NOTE: Not using controller.now.
            curr_time = self.controller.store.now_tz_aware()
        new_time = curr_time + delta_time

        if start_or_end == 'start':
            if edit_fact.end and new_time > edit_fact.end:
                new_time = edit_fact.end
        else:
            self.controller.affirm(start_or_end == 'end')
            if new_time < edit_fact.start:
                # We apply minimum Fact width (fact_min_delta) next.
                new_time = edit_fact.start
        self.edit_time_apply_time(
            edit_fact, new_time, neighbor, start_or_end, gap_okay,
        )

    # ***

    def edit_time_apply_time(
        self, edit_fact, new_time, neighbor, start_or_end, gap_okay=False,
    ):
        new_time = self.edit_time_check_edge(new_time, neighbor, start_or_end, gap_okay)
        setattr(edit_fact, start_or_end, new_time)

    # ***

    def edit_time_check_edge(self, new_time, neighbor, start_or_end, gap_okay=False):
        if neighbor is None:
            return new_time

        # FIXME/2020-04-12: (lb): Split fact_min_delta in two? There's the
        # "this is minimum width to save Fact", which I (personally) want
        # 0 (or at least < 1 second). But then there's also this version,
        # "do not automatically squash an out-of-view neighbor Fact to 0!"
        # Let's at least ensure a 1 second minimum squash neighbor width.
        fact_min_delta = 0
        if 'interval-gap' not in neighbor.dirty_reasons:
            fact_min_delta = int(self.controller.config['time.fact_min_delta'])
            fact_min_delta = max(fact_min_delta, 1)
        min_delta = timedelta(minutes=fact_min_delta)

        if start_or_end == 'start':
            end_boundary = neighbor.start + min_delta
            if neighbor.end and end_boundary > neighbor.end:
                # Ignore min_delta if time window already smaller.
                end_boundary = neighbor.end
            if new_time < end_boundary:
                new_time = end_boundary
            if not gap_okay or new_time < neighbor.end:
                neighbor.end = new_time
        else:
            self.controller.affirm(start_or_end == 'end')
            if neighbor.end:
                start_boundary = neighbor.end - min_delta
                if start_boundary < neighbor.start:
                    # Ignore min_delta if time window already smaller.
                    start_boundary = neighbor.start
                if new_time > start_boundary:
                    new_time = start_boundary
            if not gap_okay or new_time > neighbor.start:
                neighbor.start = new_time

        return new_time




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

import re
import time
from datetime import timedelta

from gettext import gettext as _

from .exceptions import catch_action_exception
from .zone_content import ZoneContent

__all__ = (
    'UpdateHandler',
)


class UpdateHandler(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel
        self.long_press_multiplier_init()
        self.count_modifier_init()

    def long_press_multiplier_init(self):
        self.last_time_time_adjust = None
        self.init_time_time_adjust = None
        self.press_cnt_time_adjust = 0

    def standup(self):
        self.edits_manager = self.carousel.edits_manager
        self.zone_manager = self.carousel.zone_manager

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def undo_command(self, event):
        """"""
        undone = self.edits_manager.undo_last_edit()
        if not undone:
            hot_notif = _("Already at oldest change")  # Like Vim says!
        else:
            self.zone_manager.rebuild_viewable()
            # FIXME: Do not exceed terminal width.
            #  hot_notif = _("Restored: {0}").format(edited_fact.short)
            hot_notif = _("Undid last change. Press Ctrl-y to redo")
        self.zone_manager.update_status(hot_notif)

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def redo_command(self, event):
        """"""
        redone = self.edits_manager.redo_last_undo()
        if not redone:
            hot_notif = _("Already at newest change")  # Same says Vim.
        else:
            self.zone_manager.rebuild_viewable()
            # FIXME: Do not exceed terminal width.
            #  hot_notif = _("Restored: {0}").format(edited_fact.short)
            hot_notif = _("Redid last change. Press Ctrl-z to undo")
        self.zone_manager.update_status(hot_notif)

    # ***

    if False:  # FIXME/2020-04-11: Implement or get off the pot!

        @catch_action_exception
        @ZoneContent.Decorators.reset_showing_help
        def fact_split(self, event):
            """"""
            self.carousel.controller.affirm(False)  # FIXME: Implement.

            self.zone_manager.rebuild_viewable()
            pass  # FIXME: Implement
            # FIXME/2019-01-15 13:51: For undo/redo, you'll need to leave
            # copy of new split Fact with copies of original Fact's
            # prev and next fact pointers, so that wires do not get crossed.

        @catch_action_exception
        @ZoneContent.Decorators.reset_showing_help
        def fact_erase(self, event):
            """"""
            # FIXME: Implement with undo, or provide other way to recover.
            self.carousel.controller.affirm(False)  # FIXME: Implement.
            self.zone_manager.rebuild_viewable()
            pass  # FIXME: Implement

        @catch_action_exception
        @ZoneContent.Decorators.reset_showing_help
        def fact_cut(self, event):
            """"""
            self.carousel.controller.affirm(False)  # FIXME: Implement.
            self.zone_manager.rebuild_viewable()
            pass  # FIXME: Implement

        @catch_action_exception
        @ZoneContent.Decorators.reset_showing_help
        def fact_merge_prev(self, event):
            """"""
            self.carousel.controller.affirm(False)  # FIXME: Implement.
            self.zone_manager.rebuild_viewable()
            pass  # FIXME: Implement

        @catch_action_exception
        @ZoneContent.Decorators.reset_showing_help
        def fact_merge_next(self, event):
            """"""
            self.carousel.controller.affirm(False)  # FIXME: Implement.
            self.zone_manager.rebuild_viewable()
            pass  # FIXME: Implement

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_activity(self, event):
        """"""
        self.edits_manager.fact_copy_activity()
        self.zone_manager.update_status(_("Copied activity"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_tags(self, event):
        """"""
        self.edits_manager.fact_copy_tags()
        self.zone_manager.update_status(_("Copied tags"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_description(self, event):
        """"""
        self.edits_manager.fact_copy_description()
        self.zone_manager.update_status(_("Copied description"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_fact(self, event):
        """"""
        self.edits_manager.fact_copy_fact()
        self.zone_manager.update_status(_("Copied fact"))

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_paste(self, event):
        """"""
        def _fact_paste():
            paste_what = self.edits_manager.paste_copied_meta()
            hot_notif = paste_response(paste_what)
            self.zone_manager.update_status(hot_notif)

        def paste_response(paste_what):
            if not paste_what:
                return _("Nothing copied, nothing pasted")
            else:
                self.zone_manager.rebuild_viewable()
                return _("Pasted {0}").format(_(paste_what))

        _fact_paste()

    # ***

    # FIXME: (lb): These time +/- bindings are pretty cool,
    #        but they should and/or create conflicts,
    #        or affect the time of adjacent facts...
    #        so you may need to show times of adjacent facts,
    #        and/or you could auto-adjust adjacent facts' times,
    #        perhaps not letting user decrement below time of fact
    #        before it, or increment above time of fact after it.
    #        (at least need to preload the before and after facts for
    #        the current fact).
    #        for now, you can raw-edit factoid to change type
    #        and resolve conflicts. or you could edit facts in
    #        carousel and resolve conflicts yourself, before saving.
    #
    #        maybe if you begin to edit time, then you show
    #        adjacent fact's (or facts') time(s).
    #
    #        I think easiest option is to just adjust adjacent
    #        facts' times -- without showing them -- and to
    #        stop at boundary, i.e., when adjacent fact would
    #        be at 0 time (albeit, when if seconds don't match?
    #        you should normalize and remove seconds when you
    #        you adjust time!)
    #    1.  - normalize to 0 seconds on adjust
    #    2.  - change adjacent facts' time(s), too
    #    3.  - stop at adjacent fact boundary
    #    4.  - option to delete fact -- so user could either change to
    #          adjacent fact and edit its time to make more room for
    #          fact being re-timed; or user could switch to adjacent
    #          fact and delete it, poof!
    #    5.  - undo/redo stack of edit_facts
    #          e.g., undo delete fact command, or undo any edit:
    #             act@cat, tags, description!
    #             keep a copy of undo fact in carousel
    #    6.  - insert new fact -- or split fact! split fact in twain...
    #          YES! you could even have a split fact command, and then
    #          a clear fact command ... (s)plit fact... (C)lear fact...
    #          or maybe (S) and (C) ... this makes the redo/undo trickier...
    #          perhaps each redo/undo state is a copy of edit_facts?
    #          i think the trickier part is coordinating new_facts
    #          and edit_facts -- the lookup can return multiple facts,
    #          i'm guessing... and then the new split facts will just
    #          show the diff each against the same base fact... and on
    #          save... what? they all have the same split_from, I guess...
    #          because we don't really use split_from yet... OK, this is
    #          doable, super doable... and it could make my fact entering
    #          at end of day easier? or fact cleanup, i should mean, e.g.,
    #          if I had a fact that was 6 hours long, I could split it
    #          in two, and then adjust the time quickly with arrow keys!
    #    7.  - (h)elp option? TOGGLE HELP IN DESCRIPTION BOX!! You could
    #          probably easily redraw the header, too, to not show a fact...
    #          maybe use Header and Description area -- can I use a new
    #          Layout item or whatever from PPT and swap out current
    #          display? OR, leave the Fact header, because user might
    #          want to use one of the HELP KEYS while VIEWING THE HELP.
    #          So show help, and on any command, clear the help!
    #          And does this mean I should add more meta to the key bindings
    #          and build the help automatically? Or should I, say, just
    #          make a custom string and maintain the help separately?
    #          the latter is probably easiest...
    #          - Use different background color when displaying the help!
    #
    #        - user can use left/right to see adjacent facts, like normal
    #          (i.e., do not add more widgets/info to carousel!)
    #        - user can adjust adjacent facts' times as well, to
    #          keep pushing facts' times around
    #
    #    X.  - swap 2 adjacent Facts? seems like it makes sense,
    #          but also seems useless, as in, what use case would
    #          have user wanting to swap 2 facts?? i've never wanted
    #          to do that.

    @catch_action_exception
    def edit_time_decrement_start(self, event):
        self.edit_time_adjust(-1, 'start')

    @catch_action_exception
    def edit_time_increment_start(self, event):
        self.edit_time_adjust(1, 'start')

    @catch_action_exception
    def edit_time_decrement_end(self, event):
        self.edit_time_adjust(-1, 'end')

    @catch_action_exception
    def edit_time_increment_end(self, event):
        self.edit_time_adjust(1, 'end')

    @catch_action_exception
    def edit_time_decrement_both(self, event):
        self.edit_time_adjust(-1, 'start', 'end')

    @catch_action_exception
    def edit_time_increment_both(self, event):
        self.edit_time_adjust(1, 'start', 'end')

    @catch_action_exception
    def edit_time_decrement_start_5min(self, event):
        self.edit_time_adjust(-5, 'start')

    @catch_action_exception
    def edit_time_increment_start_5min(self, event):
        self.edit_time_adjust(5, 'start')

    @catch_action_exception
    def edit_time_decrement_end_5min(self, event):
        self.edit_time_adjust(-5, 'end')

    @catch_action_exception
    def edit_time_increment_end_5min(self, event):
        self.edit_time_adjust(5, 'end')

    def edit_time_adjust(self, delta_mins, start_or_end, end_maybe=None):
        modifier = self.count_modifier_parse()
        if modifier is not None:
            delta_mins *= modifier
        delta_time = self.edit_time_multiplier(delta_mins)
        self.edits_manager.edit_time_adjust(
            delta_time,
            start_or_end,
            end_maybe,
            gap_okay=self.time_gap_allowed,
        )
        self.edit_time_reset_refresh()

    def edit_time_reset_refresh(self):
        self.count_modifier_reset()
        self.zone_manager.reset_diff_fact()
        self.zone_manager.selectively_refresh()

    # Hahaha, (lb): The time_adjust multiplier is a somewhat ridiculous feature.
    #   I'm not sure how useful it is.
    #   Especially if we still need a way for user to edit time directly!

    # (lb): If I hold down a key, I see about a ~ 33 msec. repeat.
    #   Which is probably "hardware settings"-settable per distro.
    #   If I repeatedly manually press a key, the fastest I can do
    #   is about 60 msecs. between presses (though it's more like
    #   100 msec. normally). So we'll try a threshold of 50 msecs.
    #   (Of course, 0.05 doesn't always work. Trying 100 msecs....)
    # 2020-04-12: Possibly because logging, I'm seeing ~0.16 delays on long
    # press. But I added reset_time_multipliers decorator to key_action_map
    # handlers, so we probably do not even need this threshold, because now
    # if last_time_time_adjust is set, it should mean user has not pressed
    # any other key since (see also count_modifier_any_key).
    DOUBLE_KEYLICK_THRESHOLD = 0.33
    MULTIPLY_BY_05_THRESHOLD = 4
    MULTIPLY_BY_10_THRESHOLD = 7
    MULTIPLY_BY_30_THRESHOLD = 9.5
    MULTIPLY_BY_60_THRESHOLD = 12
    MULTIPLY_BY_1440_THRESHOLD = 15
    MULTIPLY_BY_10080_THRESHOLD = 20
    MULTIPLY_BY_525960_THRESHOLD = 30

    def edit_time_multiplier(self, delta_mins):
        # Calculate time since last key press was processed.
        press_elapsed = 0
        if self.last_time_time_adjust:
            press_elapsed = time.time() - self.last_time_time_adjust

        # Calculate time since first key press was detected.
        total_elapsed = 0
        if self.init_time_time_adjust:
            total_elapsed = time.time() - self.init_time_time_adjust

        if (
            self.last_time_time_adjust is None
            # Be aware of the OS hardware delay before a long press becomes a
            # repeat event, and ignore elapsed time before 1st and 2nd events.
            or (
                self.press_cnt_time_adjust > 1
                and press_elapsed > UpdateHandler.DOUBLE_KEYLICK_THRESHOLD
            )
        ):
            self.init_time_time_adjust = time.time()
            self.press_cnt_time_adjust = 0
            delta_time = timedelta(minutes=delta_mins * 1)
        else:
            if total_elapsed < UpdateHandler.MULTIPLY_BY_05_THRESHOLD:
                delta_time = timedelta(minutes=delta_mins * 1)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_10_THRESHOLD:
                delta_time = timedelta(minutes=delta_mins * 5)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_30_THRESHOLD:
                delta_time = timedelta(minutes=delta_mins * 10)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_60_THRESHOLD:
                delta_time = timedelta(minutes=delta_mins * 30)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_1440_THRESHOLD:
                # 1 day == 1440 minutes.
                delta_time = timedelta(days=delta_mins)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_10080_THRESHOLD:
                # 1 week == 10080 minutes.
                delta_time = timedelta(days=delta_mins * 7)
            elif total_elapsed < UpdateHandler.MULTIPLY_BY_525960_THRESHOLD:
                # 1 year =~ 525960 minutes.
                # 365.25 days / Julian year. 365.242189 days / Tropical year.
                # 365.2425 days / Mean year. 365.25636 days / Sidereal year.
                delta_time = timedelta(days=delta_mins * 365.25636)
            else:
                delta_time = timedelta(minutes=delta_mins * 60 * 8760)
        self.last_time_time_adjust = time.time()
        self.press_cnt_time_adjust += 1

        return delta_time

    # ***

    @catch_action_exception
    def edit_fact(self, event):
        """"""
        self.carousel.enduring_edit = True
        event.app.exit()

    def exit_to_awesome_prompt(self, event, restrict_edit):
        self.carousel.enduring_edit = True
        self.carousel.restrict_edit = restrict_edit
        event.app.exit()

    @catch_action_exception
    def edit_actegory(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'actegory')

    @catch_action_exception
    def edit_tags(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'tags')

    @catch_action_exception
    def edit_description(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'description')

    # ***

    def _key_sequence_str(self, event):
        # For normal characters, we could grab event.data, but for combos,
        # use key_sequence for readability. E.g., if combo is ('X', 'c-c'),
        # event.data == '\x03', but joining key_sequences gives 'Xc-c'.
        return ''.join(seq.key for seq in event.key_sequence)

    @catch_action_exception
    def begin_commando(self, event):
        """"""
        self.began_commando = self._key_sequence_str(event)
        self.typed_commando = ''
        self.zone_manager.zone_lowdown.update_status(
            self.began_commando,
            clear_after_secs=0,
        )
        self.carousel.action_manager.wire_keys_commando()

    @catch_action_exception
    def parts_commando(self, event):
        """"""
        self.typed_commando += self._key_sequence_str(event)
        self.zone_manager.zone_lowdown.update_status(
            self.began_commando + self.typed_commando,
            clear_after_secs=0,
        )

    @catch_action_exception
    def final_commando(self, event):
        """"""
        hot_notif = self.colon_commando(event, self.typed_commando)
        del self.began_commando
        del self.typed_commando
        if hot_notif:
            self.zone_manager.zone_lowdown.update_status(hot_notif)
        else:
            self.zone_manager.zone_lowdown.reset_status()
        self.carousel.action_manager.unwire_keys_commando()

    # ***

    def colon_commando(self, event, typed_commando):
        if not typed_commando:
            return
        keys_config = self.carousel.controller.config['editor-keys']
        if typed_commando == keys_config['write_commando']:
            # E.g., `:w`.
            self.carousel.save_edited_and_live(event)
        elif typed_commando == keys_config['quit_commando']:
            # E.g., `:q`.
            self.carousel.exit_command(event)
        elif typed_commando == keys_config['save_quit_commando']:
            self.carousel.save_edited_and_exit(event)
        else:
            # (lb): Copying Vim's message for now. Verbatim. Don't judge.
            msg = 'E492: Not an editor command: {}'.format(typed_commando)
            return msg

    # ***

    def count_modifier_init(self):
        self.count_modifier = None
        self.final_modifier = None
        self.time_gap_allowed = None
        self.delta_time_target = None

    def count_modifier_setup(self):
        self.count_modifier = ''
        self.final_modifier = None
        self.time_gap_allowed = False
        # If user presses '+'/'-', then we'll set self.delta_time_target.

    # ***

    @catch_action_exception
    def count_modifier_any_key(self, event=None):
        # Note: This only catches keys not mapped elsewhere.
        # And we only care about digits and decimal points.
        if self.count_modifier is None:
            # Do not reset count modifier whilst building it!
            self.count_modifier_setup()
        self.count_modifier_feed(event)

    # ***

    @catch_action_exception
    def begin_delta_time_start(self, event):
        """"""
        self.carousel.action_manager.wire_keys_delta_time()
        if self.count_modifier is None:
            # This allows user to press '!' before '+'.
            self.count_modifier_setup()
        self.delta_time_target = 'end'

    @catch_action_exception
    def begin_delta_time_end(self, event):
        """"""
        self.carousel.action_manager.wire_keys_delta_time()
        if self.count_modifier is None:
            # This allows user to press '!' before '-'.
            self.count_modifier_setup()
        self.delta_time_target = 'start'

    RE_NUMERIC = re.compile('^[0-9]$')

    @catch_action_exception
    def parts_delta_time(self, event):
        """"""
        self.count_modifier_feed(event)

    def count_modifier_feed(self, event):
        # If user pressed allow-gap key already (e.g., '!'), do not allow
        # more time to be entered; be strict and reset instead. Unless
        # user has not entered any time yet.
        if self.final_modifier:
            self.count_modifier_reset()
        # We could allow any character and then parse at error, potentially
        # showing an error message. Or, if user types non-number now, we
        # could just go back to old state. -- It's not completely silent,
        # as the cursor will reappear.
        elif event.data == '.':
            if '.' in self.count_modifier:
                # So strict!
                self.count_modifier_reset()
            else:
                self.count_modifier += '.'
        elif UpdateHandler.RE_NUMERIC.match(event.data):
            self.count_modifier += event.data
        else:
            # Not '[0-9]' or '.'. Ignore it and reset state.
            self.count_modifier_reset()

    @catch_action_exception
    def allow_time_gap(self, event):
        """"""
        if self.time_gap_allowed is True:
            # Be strict. Consider a second '!' to be syntax error, so reset.
            self.count_modifier_reset()
        else:
            if self.count_modifier is None:
                # Allow '!' to precede '[0-9]' or '+'/'-'.
                self.count_modifier_setup()
            elif self.count_modifier:
                # User entered some numbers or a decimal already. Mark it final
                self.final_modifier = self.count_modifier
            self.time_gap_allowed = True

    @catch_action_exception
    def final_delta_time_apply(self, event):
        """"""
        self.apply_delta_time(event, scalar=1)

    @catch_action_exception
    def final_delta_time_minutes(self, event):
        """"""
        self.apply_delta_time(event, scalar=1)

    @catch_action_exception
    def final_delta_time_hours(self, event):
        """"""
        self.apply_delta_time(event, scalar=60)

    def apply_delta_time(self, event, scalar):
        delta_time = self.count_modifier_parse()
        if delta_time is None:
            # Happens if user types, e.g., `+h`, i.e., without any digits.
            self.edit_time_reset_refresh()
            return
        delta_mins = timedelta(minutes=delta_time * scalar)
        edit_fact = self.edits_manager.editable_fact()
        if self.delta_time_target == 'end':
            apply_time = edit_fact.start + delta_mins
        else:  # == 'start'
            end_time = edit_fact.end or self.carousel.controller.now
            apply_time = end_time - delta_mins

        self.edits_manager.edit_time_adjust(
            apply_time,
            self.delta_time_target,
            gap_okay=self.time_gap_allowed,
        )
        self.edit_time_reset_refresh()

    def count_modifier_parse(self):
        if not self.count_modifier:
            return None
        try:
            delta_time = int(self.count_modifier)
        except ValueError:
            delta_time = float(self.count_modifier)
        return delta_time

    def apply_count_multiplier(self, count=1, floats=False):
        modifier = self.count_modifier_parse()
        if modifier is not None:
            count = count * modifier
            if not floats:
                count = int(count)
        self.reset_time_multipliers()
        return count

    @catch_action_exception
    def panic_delta_time(self, event):
        """"""
        self.count_modifier_reset()

    def reset_time_multipliers(self):
        self.count_modifier_reset()
        # Also the long-press multipliers.
        self.last_time_time_adjust = None
        self.init_time_time_adjust = None
        self.press_cnt_time_adjust = 0

    def count_modifier_reset(self):
        if self.count_modifier is None:
            return
        self.count_modifier = None
        self.final_modifier = None
        self.time_gap_allowed = None
        if self.delta_time_target is not None:
            self.carousel.action_manager.unwire_keys_delta_time()
            self.delta_time_target = None

